"""
pygame-menu
https://github.com/ppizarror/pygame-menu

UTILS
Utility functions.
"""

__all__ = [

    # Methods
    'assert_alignment',
    'assert_color',
    'assert_cursor',
    'assert_list_vector',
    'assert_orientation',
    'assert_position',
    'assert_position_vector',
    'assert_vector',
    'check_key_pressed_valid',
    'fill_gradient',
    'format_color',
    'get_cursor',
    'get_finger_pos',
    'load_pygame_image_file',
    'make_surface',
    'mouse_motion_current_mouse_position',
    'parse_padding',
    'print_menu_widget_structure',
    'set_pygame_cursor',
    'uuid4',
    'warn',
    'widget_terminal_title',

    # Constants
    'PYGAME_V2',

    # Classes
    'ShadowGenerator',
    'TerminalColors'

]

import sys
import traceback
import uuid
import warnings

import pygame
import pygame_menu

from pygame_menu.locals import ALIGN_CENTER, ALIGN_LEFT, ALIGN_RIGHT, POSITION_CENTER, \
    POSITION_NORTH, POSITION_SOUTH, POSITION_SOUTHEAST, POSITION_NORTHWEST, \
    POSITION_WEST, POSITION_EAST, POSITION_NORTHEAST, POSITION_SOUTHWEST, \
    ORIENTATION_HORIZONTAL, ORIENTATION_VERTICAL, FINGERDOWN, FINGERUP, FINGERMOTION

from pygame_menu._types import ColorType, ColorInputType, Union, List, Vector2NumberType, \
    NumberType, Any, Optional, Tuple, NumberInstance, VectorInstance, PaddingInstance, \
    PaddingType, Tuple4IntType, ColorInputInstance, VectorType, EventType, \
    CursorInputInstance, CursorInputType, Tuple2IntType, Dict, Tuple3IntType

PYGAME_V2 = pygame.version.vernum[0] >= 2
WARNINGS_LAST_MESSAGES: Dict[int, bool] = {}


def assert_alignment(align: str) -> None:
    """
    Assert that a certain alignment is valid.

    :param align: Align value
    """
    assert isinstance(align, str), f'alignment "{align}" must be a string'
    assert align in (ALIGN_LEFT, ALIGN_CENTER, ALIGN_RIGHT), \
        f'incorrect alignment value "{align}"'


def assert_color(
    color: Union[ColorInputType, List[int]],
    warn_if_invalid: bool = True
) -> ColorType:
    """
    Assert that a certain color is valid.

    :param color: Object color
    :param warn_if_invalid: If ``True`` warns if the color is invalid
    :return: Formatted color if valid, else, throws an ``AssertionError`` exception
    """
    color = format_color(color, warn_if_invalid=warn_if_invalid)
    assert isinstance(color, VectorInstance), \
        f'color must be a tuple or list, not type "{type(color)}"'
    assert 4 >= len(color) >= 3, \
        'color must be a tuple or list of 3 or 4 numbers'
    for i in range(3):
        assert isinstance(color[i], int), \
            f'"{color[i]}" in element color {color} must be an integer, not type "{type(color)}"'
        assert 0 <= color[i] <= 255, \
            f'"{color[i]}" in element color {color} must be an integer between 0 and 255'
    if len(color) == 4:
        assert isinstance(color[3], int), \
            f'alpha channel must be an integer between 0 and 255, not type "{type(color)}"'
        assert 0 <= color[3] <= 255, \
            f'opacity of color {color} must be an integer between 0 and 255; ' \
            f'where 0 is fully-transparent and 255 is fully-opaque'
    return color


def assert_cursor(cursor: CursorInputType) -> None:
    """
    Assert a given cursor is valid.

    :param cursor: Cursor object
    """
    assert isinstance(cursor, CursorInputInstance), \
        'cursor instance invalid, it can be None, an integer, ' \
        'or pygame.cursors.Cursor'


def assert_list_vector(list_vector: Union[List[Vector2NumberType], Tuple[Vector2NumberType, ...]],
                       length: int) -> None:
    """
    Assert that a list fixed length vector is numeric.

    :param list_vector: Numeric list vector
    :param length: Length of the required vector. If ``0`` don't check the length
    """
    assert isinstance(list_vector, VectorInstance), \
        f'list_vector "{list_vector}" must be a tuple or list'
    for v in list_vector:
        assert_vector(v, length)


def assert_orientation(orientation: str) -> None:
    """
    Assert that a certain widget orientation is valid.

    :param orientation: Object orientation
    """
    assert isinstance(orientation, str), \
        f'orientation "{orientation}" must be a string'
    assert orientation in (ORIENTATION_HORIZONTAL, ORIENTATION_VERTICAL), \
        f'invalid orientation value "{orientation}"'


def assert_position(position: str) -> None:
    """
    Assert that a certain position is valid.

    :param position: Object position
    """
    assert isinstance(position, str), \
        f'position "{position}" must be a string'
    assert position in (POSITION_WEST, POSITION_SOUTHWEST, POSITION_SOUTH,
                        POSITION_SOUTHEAST, POSITION_EAST, POSITION_NORTH,
                        POSITION_NORTHWEST, POSITION_NORTHEAST, POSITION_CENTER), \
        f'invalid position value "{position}"'


def assert_position_vector(position: Union[str, List[str], Tuple[str, ...]]) -> None:
    """
    Assert that a position vector is valid.

    :param position: Object position
    """
    if isinstance(position, str):
        assert_position(position)
    else:
        assert isinstance(position, VectorInstance)
        unique = []
        for pos in position:
            assert_position(pos)
            if pos not in unique:
                unique.append(pos)
        assert len(unique) == len(position), 'there cannot be repeated positions'


def assert_vector(
    num_vector: VectorType,
    length: int,
    instance: type = NumberInstance
) -> None:
    """
    Assert that a fixed length vector is numeric.

    :param num_vector: Numeric vector
    :param length: Length of the required vector. If ``0`` don't check the length
    :param instance: Instance of each item of the vector
    """
    assert isinstance(num_vector, VectorInstance), \
        f'vector "{num_vector}" must be a list or tuple of {length} items if type {instance}'
    if length != 0:
        assert len(num_vector) == length, \
            f'vector "{num_vector}" must contain {length} numbers only, ' \
            f'but {num_vector} were given'
    for i in range(len(num_vector)):
        num = num_vector[i]
        if instance == int and isinstance(num, float) and int(num) == num:
            num = int(num)
        assert isinstance(num, instance), \
            f'item {num} of vector must be {instance}, not type "{type(num)}"'


def check_key_pressed_valid(event: EventType) -> bool:
    """
    Checks if the pressed key is valid.

    :param event: Key press event
    :return: ``True`` if a key is pressed
    """
    # If the system detects that any key event has been pressed but
    # there's not any key pressed then this method raises a KEYUP
    # flag
    bad_event = not (True in pygame.key.get_pressed())
    if bad_event:
        if 'test' in event.dict and event.dict['test']:
            return True
        ev = pygame.event.Event(pygame.KEYUP, {'key': event.key})
        pygame.event.post(ev)
    return not bad_event


def fill_gradient(
    surface: 'pygame.Surface',
    color: ColorInputType,
    gradient: ColorInputType,
    rect: Optional['pygame.Rect'] = None,
    vertical: bool = True,
    forward: bool = True
) -> None:
    """
    Fill a surface with a gradient pattern.

    :param surface: Surface to fill
    :param color: Starting color
    :param gradient: Final color
    :param rect: Area to fill; default is surface's rect
    :param vertical: True=vertical; False=horizontal
    :param forward: True=forward; False=reverse
    """
    if rect is None:
        rect = surface.get_rect()
    x1, x2 = rect.left, rect.right
    y1, y2 = rect.top, rect.bottom
    color = assert_color(color)
    gradient = assert_color(gradient)
    if vertical:
        h = y2 - y1
    else:
        h = x2 - x1
    if forward:
        a, b = color, gradient
    else:
        b, a = color, gradient
    rate = (
        float(b[0] - a[0]) / h,
        float(b[1] - a[1]) / h,
        float(b[2] - a[2]) / h
    )
    fn_line = pygame.draw.line
    if vertical:
        for line in range(y1, y2):
            color = (
                min(max(a[0] + (rate[0] * (line - y1)), 0), 255),
                min(max(a[1] + (rate[1] * (line - y1)), 0), 255),
                min(max(a[2] + (rate[2] * (line - y1)), 0), 255)
            )
            fn_line(surface, color, (x1, line), (x2, line))
    else:
        for col in range(x1, x2):
            color = (
                min(max(a[0] + (rate[0] * (col - x1)), 0), 255),
                min(max(a[1] + (rate[1] * (col - x1)), 0), 255),
                min(max(a[2] + (rate[2] * (col - x1)), 0), 255)
            )
            fn_line(surface, color, (col, y1), (col, y2))


def format_color(
    color: Union[ColorInputType, Any],
    warn_if_invalid: bool = True
) -> Union[ColorType, Any]:
    """
    Format color from string, int, or tuple to tuple type.

    Available formats:
    - Color name str: name of the color to use, e.g. ``"red"`` (all the supported name strings can be found in the colordict module, see https://github.com/pygame/pygame/blob/main/src_py/colordict.py)
    - HTML color format str: ``"#rrggbbaa"`` or ``"#rrggbb"``, where rr, gg, bb, and aa are 2-digit hex numbers in the range of ``0`` to ``0xFF`` inclusive, the aa (alpha) value defaults to ``0xFF`` if not provided
    - Hex number str: ``"0xrrggbbaa"`` or ``"0xrrggbb"``, where rr, gg, bb, and aa are 2-digit hex numbers in the range of ``0x00`` to ``0xFF`` inclusive, the aa (alpha) value defaults to ``0xFF`` if not provided
    - int: int value of the color to use, using hex numbers can make this parameter more readable, e.g. ``0xrrggbbaa``, where rr, gg, bb, and aa are 2-digit hex numbers in the range of ``0x00`` to ``0xFF`` inclusive, note that the aa (alpha) value is not optional for the int format and must be provided
    - tuple/list of int color values: ``(R, G, B, A)`` or ``(R, G, B)``, where R, G, B, and A are int values in the range of ``0`` to ``255`` inclusive, the A (alpha) value defaults to ``255`` (opaque) if not provided

    :param color: Color to format. If format is valid returns the same input value
    :param warn_if_invalid: If ``True`` warns if the color is invalid
    :return: Color in (r, g, b, a) format
    """
    if not isinstance(color, ColorInputInstance):
        return color
    if not isinstance(color, pygame.Color):
        if isinstance(color, str):
            if len(color) == 4 and color[0] == '#':
                r, g, b = color[1], color[2], color[3]
                color = f'#{r * 2}{g * 2}{b * 2}'
        try:
            if isinstance(color, VectorInstance) and 3 <= len(color) <= 4:
                if PYGAME_V2:
                    for j in color:
                        if not isinstance(j, int):
                            raise ValueError('color cannot contain floating point values')
                c = pygame.Color(*color)
            else:
                c = pygame.Color(color)
        except ValueError:
            if warn_if_invalid:
                warn(f'invalid color value "{color}"')
            else:
                raise
            return color
    else:
        c = color
    return c.r, c.g, c.b, c.a


def get_cursor() -> CursorInputType:
    """
    Return the pygame cursor object.

    :return: Cursor object
    """
    try:
        return pygame.mouse.get_cursor()
    except TypeError as e:
        warn(str(e))
    return None


def get_finger_pos(menu: Optional['pygame_menu.Menu'], event: EventType) -> Tuple2IntType:
    """
    Return the position from finger (or mouse) event on x-axis and y-axis (x, y).

    :param menu: Menu object for relative positioning in finger events
    :param event: Pygame event object
    :return: Position on x-axis and y-axis (x, y) in px
    """
    if event.type in (FINGERDOWN, FINGERMOTION, FINGERUP):
        assert menu is not None, \
            'menu reference cannot be none while using finger position'
        display_size = menu.get_window_size()
        finger_pos = (int(event.x * display_size[0]), int(event.y * display_size[1]))
        return finger_pos
    return event.pos


def is_callable(func: Any) -> bool:
    """
    Return ``True`` if ``func`` is callable.

    :param func: Function object
    :return: ``True`` if function
    """
    e = 'is_callable(func) method will be removed in v5, consider using built-in' \
        ' callable(func) method instead'
    warnings.warn(e, DeprecationWarning)
    return callable(func)


def load_pygame_image_file(image_path: str, **kwargs) -> 'pygame.Surface':
    """
    Loads an image and returns a surface.

    :param image_path: Image file
    :param kwargs: Optional keyword arguments
    :return: Surface
    """
    # Try to load the image
    try:
        if 'test' in kwargs.keys():
            raise pygame.error('File is not a Windows BMP file')

        surface = pygame.image.load(image_path)

    except pygame.error as exc:
        # Check if file is not a Windows file
        if str(exc) == 'File is not a Windows BMP file':
            pil_invalid_exception = Exception

            # Check if Pillow exists
            try:
                # noinspection PyPackageRequirements
                from PIL import Image, UnidentifiedImageError

                pil_invalid_exception = UnidentifiedImageError
                img_pil = Image.open(image_path)
                # noinspection PyTypeChecker
                surface = pygame.image.fromstring(
                    img_pil.tobytes(), img_pil.size, img_pil.mode).convert()

            except (ModuleNotFoundError, ImportError):
                warn(f'Image file "{image_path}" could not be loaded, as pygame.error '
                     f'is raised. To avoid this issue install the Pillow library')
                raise

            except pil_invalid_exception:
                warn(f'The image "{image_path}" could not be loaded using Pillow')
                raise

        else:
            raise

    return surface


def make_surface(
    width: NumberType,
    height: NumberType,
    alpha: bool = False,
    fill_color: Optional[ColorInputType] = None
) -> 'pygame.Surface':
    """
    Creates a pygame surface object.

    :param width: Surface width
    :param height: Surface height
    :param alpha: Enable alpha channel on surface
    :param fill_color: Fill surface with a certain color
    :return: Pygame surface
    """
    assert isinstance(width, NumberInstance)
    assert isinstance(height, NumberInstance)
    assert isinstance(alpha, bool)
    assert width >= 0 and height >= 0, \
        'surface width and height must be equal or greater than zero'
    surface = pygame.Surface((int(width), int(height)), pygame.SRCALPHA, 32)
    if alpha:
        # noinspection PyArgumentList
        surface = pygame.Surface.convert_alpha(surface)
    if fill_color is not None:
        fill_color = assert_color(fill_color)
        surface.fill(fill_color)
    return surface


def mouse_motion_current_mouse_position() -> EventType:
    """
    Return a pygame event type MOUSEMOTION in the current mouse position.

    :return: Event
    """
    x, y = pygame.mouse.get_pos()
    return pygame.event.Event(pygame.MOUSEMOTION, {'pos': (int(x), int(y))})


def parse_padding(padding: PaddingType) -> Tuple4IntType:
    """
    Get the padding value from tuple.

    - If an integer or float is provided: top, right, bottom and left values will be the same
    - If 2-item tuple is provided: top and bottom takes the first value, left and right the second
    - If 3-item tuple is provided: top will take the first value, left and right the second, and bottom the third
    - If 4-item tuple is provided: padding will be (top, right, bottom, left)

    .. note::

        See `CSS W3Schools <https://www.w3schools.com/css/css_padding.asp>`_ for more info about padding.

    :param padding: Can be a single number, or a tuple of 2, 3 or 4 elements following CSS style
    :return: Padding value, (top, right, bottom, left), in px
    """
    if padding is False or None:
        padding = 0
    assert isinstance(padding, PaddingInstance)

    if isinstance(padding, NumberInstance):
        assert padding >= 0, 'padding cannot be a negative number'
        return int(padding), int(padding), int(padding), int(padding)
    else:
        assert 1 <= len(padding) <= 4, 'padding must be a tuple of 2, 3 or 4 elements'
        for i in range(len(padding)):
            assert isinstance(padding[i], NumberInstance), \
                'all padding elements must be integers or floats'
            assert padding[i] >= 0, \
                'all padding elements must be equal or greater than zero'
        if len(padding) == 1:
            return int(padding[0]), int(padding[0]), int(padding[0]), int(padding[0])
        elif len(padding) == 2:
            return int(padding[0]), int(padding[1]), int(padding[0]), int(padding[1])
        elif len(padding) == 3:
            return int(padding[0]), int(padding[1]), int(padding[2]), int(padding[1])
        else:
            return int(padding[0]), int(padding[1]), int(padding[2]), int(padding[3])


def print_menu_widget_structure(
    widgets: List['pygame_menu.widgets.Widget'],
    index: int
) -> None:
    """
    Test printing widgets order.

    .. note::

        - Φ       Floating status
        - ⇇      Selected
        - !▲      Widget is not appended to current menu
        - ╳      Widget is hidden
        - ∑       Scrollable frame sizing
        - β       Widget is not selectable
        - {x,y}   Widget *column, row* position
        - <x,y>   Frame indices (min, max)

    :param widgets: Menu widgets list
    :param index: Menu index
    """
    indx = 0
    current_depth = 0
    depth_widths = {}
    c = TerminalColors

    def close_frames(depth: int) -> None:
        """
        Close frames up to current depth.

        :param depth: Depth to close
        """
        d = current_depth - depth
        for i in range(d):
            j = depth + d - (i + 1)  # Current depth
            line = f'·   {"│   " * j}└{"┄" * 3}'  # * depth_widths[j]
            print(c.BRIGHT_WHITE + line.ljust(0, '━') + c.ENDC)  # 80 also work

    non_menu_frame_widgets: Dict[int, List['pygame_menu.widgets.Widget']] = {}

    def process_non_menu_frame(w_indx: int) -> None:
        """
        Print non-menu frames list.

        :param w_indx: Current iteration index to print widgets
        """
        for nmi in list(non_menu_frame_widgets.keys()):
            if nmi == w_indx:
                v = non_menu_frame_widgets[nmi]
                for v_wid in v:
                    try:
                        print(c.BRIGHT_WHITE + '·   ' + '│   ' * v_wid.get_frame_depth()
                              + c.ENDC + widget_terminal_title(v_wid))
                    except UnicodeEncodeError:
                        pass
                del non_menu_frame_widgets[nmi]

    for w in widgets:
        w_depth = w.get_frame_depth()
        close_frames(w.get_frame_depth())
        title = widget_terminal_title(w, indx, index)
        try:
            print('{0}{1}{2}'.format(
                str(indx).ljust(3),
                ' ' + c.BRIGHT_WHITE + '│   ' * w_depth + c.ENDC,
                title
            ))
        except UnicodeEncodeError:
            pass
        if w_depth not in depth_widths.keys():
            depth_widths[w_depth] = 0
        # depth_widths[w_depth] = max(int(len(title) * 1.2) + 3, depth_widths[w_depth])
        depth_widths[w_depth] = len(title) - 2
        current_depth = w.get_frame_depth()
        process_non_menu_frame(indx)
        jw = widgets[0]
        try:
            if isinstance(w, pygame_menu.widgets.Frame):  # Print ordered non-menu widgets
                current_depth += 1
                prev_indx = indx
                for jw in w.get_widgets(unpack_subframes=False):
                    if jw.get_menu() is None or jw not in widgets:
                        if prev_indx not in non_menu_frame_widgets.keys():
                            non_menu_frame_widgets[prev_indx] = []
                        non_menu_frame_widgets[prev_indx].append(jw)
                    else:
                        prev_indx = widgets.index(jw)
        except ValueError as e:
            print(f'[ERROR] while requesting widget {jw.get_class_id()}')
            warn(str(e))
        indx += 1
    process_non_menu_frame(indx)
    close_frames(0)


def set_pygame_cursor(cursor: CursorInputType) -> None:
    """
    Set pygame cursor.

    :param cursor: Cursor object
    """
    try:
        if cursor is not None:
            # noinspection PyArgumentList
            pygame.mouse.set_cursor(cursor)
    except (pygame.error, TypeError):
        if PYGAME_V2:
            warn(f'could not establish widget cursor, invalid value {cursor}')


def uuid4(short: bool = False) -> str:
    """
    Create custom version of uuid4.

    :param short: If ``True`` only returns the first 8 chars of the uuid, else, 18
    :return: UUID of 18 chars
    """
    return str(uuid.uuid4())[:18 if not short else 8]


def warn(message: str, print_stack: bool = True) -> None:
    """
    Warnings warn method.

    :param message: Message to warn about
    :param print_stack: Print stack trace of the call
    """
    assert isinstance(message, str)

    # noinspection PyUnresolvedReferences,PyProtectedMember
    frame = sys._getframe().f_back
    # frame_info = inspect.getframeinfo(frame)  # Traceback(filename, lineno, function, code_context, index)

    # Check if message in dict
    msg_hash = hash(message)
    msg_in_hash = False
    try:
        msg_in_hash = WARNINGS_LAST_MESSAGES[msg_hash]
    except KeyError:
        pass

    if not msg_in_hash and print_stack:
        traceback.print_stack(frame, limit=5)
        WARNINGS_LAST_MESSAGES[msg_hash] = True

    # warnings.showwarning(message, UserWarning, frame_info[0], frame_info[1])
    warnings.warn(message, stacklevel=2)


def widget_terminal_title(
    widget: 'pygame_menu.widgets.Widget',
    widget_index: int = -1,
    current_index: int = -1
) -> str:
    """
    Return widget title to be printed on terminals.

    :param widget: Widget to get title from
    :param widget_index: Widget index
    :param current_index: Menu index
    :return: Widget title
    """
    w_class_id = TerminalColors.BOLD + widget.get_class_id() + TerminalColors.ENDC
    if isinstance(widget, pygame_menu.widgets.Frame):
        w_title = TerminalColors.BRIGHT_WHITE + '┌━' + TerminalColors.ENDC
        w_title += f'{0} - {3}[{1},{2},'.format(w_class_id, *widget.get_indices(), TerminalColors.LGREEN)
        if widget.horizontal:
            w_title += 'H] '
        else:
            w_title += 'V] '
        if widget.is_scrollable:
            wsz = widget.get_inner_size()
            wsm = widget.get_max_size()
            wsh = wsm[0] if wsm[0] == wsz[0] else f'{wsm[0]}→{wsz[0]}'
            wsv = wsm[1] if wsm[1] == wsz[1] else f'{wsm[1]}→{wsz[1]}'
            w_title += f'∑ [{wsh},{wsv}] '
        w_title += TerminalColors.ENDC
    else:
        if widget.get_title() != '':
            title_f = TerminalColors.UNDERLINE + widget.get_title() + TerminalColors.ENDC
            w_title = f'{w_class_id} - {title_f} - '
        else:
            w_title = w_class_id + ' - '

    # Column/Row position
    w_title += TerminalColors.INDIGO
    cr = widget.get_col_row_index()
    w_title += '{' + str(cr[0]) + ',' + str(cr[1]) + '}'
    w_title += TerminalColors.ENDC

    # Add position
    w_title += TerminalColors.MAGENTA
    w_title += ' ({0},{1})'.format(*widget.get_position())
    w_title += TerminalColors.ENDC

    # Add size
    w_title += TerminalColors.BLUE
    w_title += ' ({0},{1})'.format(*widget.get_size())
    w_title += TerminalColors.ENDC

    # Add mods
    w_title += TerminalColors.CYAN
    if widget.is_floating():
        w_title += ' Φ'
    if not widget.is_visible():
        w_title += ' ╳'
    if not widget.is_selectable:
        w_title += ' β'
    if widget.is_selected():
        w_title += TerminalColors.BOLD + ' ⟵'
        if current_index != -1 and current_index != widget_index:
            w_title += f'! [{widget_index}->{current_index}]'
    if widget.get_menu() is None:
        w_title += ' !▲'
    w_title += TerminalColors.ENDC

    return w_title


class TerminalColors(object):
    """
    Terminal colors.

    See https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html.
    """
    BLUE = '\u001b[38;5;27m'
    BOLD = '\033[1m'
    BRIGHT_MAGENTA = '\u001b[35;1m'
    BRIGHT_WHITE = '\u001b[37;1m'
    CYAN = '\u001b[36m'
    ENDC = '\u001b[0m'
    GRAY = '\u001b[30;1m'
    INDIGO = '\u001b[38;5;129m'
    LGREEN = '\u001b[38;5;150m'
    MAGENTA = '\u001b[35m'
    RED = '\u001b[31m'
    UNDERLINE = '\033[4m'


class ShadowGenerator(object):
    """
    A class to generate surfaces that work as a 'shadow' for rectangular UI elements. Base shadow
    surface are generated with an algorithm, then when one is requested at a specific size the
    closest pre-generated shadow surface is picked and then scaled to the exact size requested.

    By default, it creates four base shadows in a small range of sizes. If you find the shadow
    appearance unsatisfactory then it is possible to create closer to the size of the
    elements you are having trouble with.

    Source: https://github.com/MyreMylar/pygame_gui with many edits.
    """

    _created_ellipse_shadows: Dict[str, 'pygame.Surface']
    _preloaded_shadow_corners: Dict[str, Dict[str, 'pygame.Surface']]
    _short_term_rect_cache: Dict[str, 'pygame.Surface']

    def __init__(self) -> None:
        self._created_ellipse_shadows = {}
        self._preloaded_shadow_corners = {}
        self._short_term_rect_cache = {}

    def clear_short_term_caches(self, force: bool = False) -> None:
        """
        Empties short term caches, so we aren't hanging on to so many surfaces.

        :param force: Force clear
        """
        t = len(self._created_ellipse_shadows) + len(self._preloaded_shadow_corners) + \
            len(self._short_term_rect_cache)
        if t >= 100 or force:
            self._created_ellipse_shadows.clear()
            self._preloaded_shadow_corners.clear()
            self._short_term_rect_cache.clear()

    def _create_shadow_corners(
        self,
        shadow_width_param: int,
        corner_radius_param: int,
        color: Tuple3IntType,
        aa_amount: int = 4
    ) -> Dict[str, 'pygame.Surface']:
        """
        Create corners for our rectangular shadows. These can be used across many
        sizes of shadow with the same shadow width and corner radius.

        :param shadow_width_param: Width of the shadow
        :param corner_radius_param: Corner radius of the shadow
        :param color: Shadow color
        :param aa_amount: Anti-aliasing amount. Defaults to 4x
        :return: Dict that contain the shadows of each border
        """
        shadow_width_param = max(1, shadow_width_param)

        corner_rect = pygame.Rect(
            0, 0,
            corner_radius_param * aa_amount,
            corner_radius_param * aa_amount
        )

        corner_surface, edge_surface = self._create_single_corner_and_edge(
            aa_amount=aa_amount,
            corner_radius_param=corner_radius_param,
            corner_rect=corner_rect,
            shadow_width_param=shadow_width_param,
            color=color
        )

        sub_radius = ((corner_radius_param - shadow_width_param) * aa_amount)
        top_edge = pygame.transform.smoothscale(edge_surface,
                                                (shadow_width_param, shadow_width_param))
        left_edge = pygame.transform.rotate(top_edge, 90)

        tl_corner = pygame.transform.smoothscale(corner_surface,
                                                 (corner_radius_param,
                                                  corner_radius_param))

        if sub_radius > 0:
            corner_sub_surface = pygame.surface.Surface(corner_rect.size,
                                                        flags=pygame.SRCALPHA,
                                                        depth=32)
            corner_sub_surface.fill(pygame.Color('#00000000'))

            pygame.draw.circle(corner_sub_surface,
                               pygame.Color('#FFFFFFFF'),
                               corner_rect.size,
                               sub_radius)

            corner_small_sub_surface = pygame.transform.smoothscale(corner_sub_surface,
                                                                    (corner_radius_param,
                                                                     corner_radius_param))

            tl_corner.blit(corner_small_sub_surface,
                           (0, 0),
                           special_flags=pygame.BLEND_RGBA_SUB)

        corners_and_edges = {
            'bottom': pygame.transform.flip(top_edge, False, True),
            'bottom_left': pygame.transform.flip(tl_corner, False, True),
            'bottom_right': pygame.transform.flip(tl_corner, True, True),
            'left': left_edge,
            'right': pygame.transform.flip(left_edge, True, False),
            'top': top_edge,
            'top_left': tl_corner,
            'top_right': pygame.transform.flip(tl_corner, True, False)
        }
        self._preloaded_shadow_corners[(str(shadow_width_param) +
                                        'x' +
                                        str(corner_radius_param))] = corners_and_edges

        return corners_and_edges

    @staticmethod
    def _create_single_corner_and_edge(
        aa_amount: int,
        corner_radius_param: int,
        corner_rect: 'pygame.Rect',
        shadow_width_param: int,
        color: Tuple3IntType
    ) -> Tuple['pygame.Surface', 'pygame.Surface']:
        """
        Creates a single corner surface and a single edge surface for a shadow.

        :param aa_amount: Amount of anti-aliasing
        :param corner_radius_param: Radius of a corner this shadow will go around
        :param corner_rect: Rectangular size of corner
        :param shadow_width_param: Width of shadow
        :param color: Shadow color
        :return: A tuple of the corner surface and the edge surface
        """
        aa_amount = max(1, aa_amount)
        final_corner_surface = pygame.surface.Surface((corner_radius_param * aa_amount,
                                                       corner_radius_param * aa_amount),
                                                      flags=pygame.SRCALPHA, depth=32)
        final_corner_surface.fill(pygame.Color('#00000000'))
        final_edge_surface = pygame.surface.Surface((shadow_width_param * aa_amount,
                                                     shadow_width_param * aa_amount),
                                                    flags=pygame.SRCALPHA, depth=32)
        final_edge_surface.fill(pygame.Color('#00000000'))
        corner_radius = corner_radius_param * aa_amount
        corner_centre = (corner_radius, corner_radius)
        edge_rect = pygame.Rect(0, 0,
                                shadow_width_param * aa_amount,
                                shadow_width_param * aa_amount)
        edge_shadow_fade_height = edge_rect.width

        alpha_increment = 20.0 / (shadow_width_param ** 1.5)
        shadow_alpha = alpha_increment
        r, g, b = color
        for _ in range(shadow_width_param):
            if corner_rect.width > 0 and corner_rect.height > 0 and corner_radius > 0:
                # Edge
                edge_shadow_surface = pygame.surface.Surface(
                    edge_rect.size,
                    flags=pygame.SRCALPHA,
                    depth=32)
                edge_shadow_surface.fill(pygame.Color('#00000000'))
                edge_shadow_surface.fill(pygame.Color(r, g, b, int(shadow_alpha)),
                                         pygame.Rect(0,
                                                     edge_rect.height - edge_shadow_fade_height,
                                                     edge_rect.width,
                                                     edge_shadow_fade_height))

                final_edge_surface.blit(edge_shadow_surface,
                                        (0, 0),
                                        special_flags=pygame.BLEND_RGBA_ADD)

                # Corner
                corner_shadow_surface = pygame.surface.Surface(corner_rect.size,
                                                               flags=pygame.SRCALPHA,
                                                               depth=32)
                corner_shadow_surface.fill(pygame.Color('#00000000'))
                pygame.draw.circle(corner_shadow_surface,
                                   pygame.Color(r, g, b, int(shadow_alpha)),
                                   corner_centre,
                                   corner_radius)

                final_corner_surface.blit(corner_shadow_surface,
                                          (0, 0),
                                          special_flags=pygame.BLEND_RGBA_ADD)

                # increments/decrements
                shadow_alpha += alpha_increment
                corner_radius -= aa_amount
                edge_shadow_fade_height -= aa_amount
        return final_corner_surface, final_edge_surface

    def create_new_rectangle_shadow(
        self,
        width: int,
        height: int,
        shadow_width_param: int,
        corner_radius_param: int,
        aa_amount: int = 4,
        color: Tuple3IntType = (0, 0, 0)
    ) -> Optional['pygame.Surface']:
        """
        Creates a rectangular shadow surface at the specified size and stores it for later use.

        :param width: The width of the base shadow to create
        :param height: The height of the base shadow to create
        :param shadow_width_param: The width of the shadowed edge
        :param corner_radius_param: The radius of the rectangular shadow's corners
        :param aa_amount: Antialiasing
        :param color: Shadow color (r, g, b)
        return: Shadow
        """
        assert isinstance(width, int)
        assert isinstance(height, int)
        assert_vector(color, 3, int)
        shadow_width_param, corner_radius_param, aa_amount = int(shadow_width_param), \
            int(corner_radius_param), int(aa_amount)
        if width < corner_radius_param or height < corner_radius_param or shadow_width_param == 0:
            return None
        r, g, b = color
        params = [width, height, shadow_width_param, corner_radius_param, aa_amount, r, g, b]
        shadow_id = '_'.join(str(param) for param in params)
        if shadow_id in self._short_term_rect_cache:
            return self._short_term_rect_cache[shadow_id]
        final_surface = pygame.surface.Surface((width, height), flags=pygame.SRCALPHA, depth=32)
        final_surface.fill(pygame.Color('#00000000'))

        corner_index_id = str(shadow_width_param) + 'x' + str(corner_radius_param)
        if corner_index_id in self._preloaded_shadow_corners:
            edges_and_corners = self._preloaded_shadow_corners[corner_index_id]
        else:
            edges_and_corners = self._create_shadow_corners(
                shadow_width_param=shadow_width_param,
                corner_radius_param=corner_radius_param,
                color=color,
                aa_amount=aa_amount
            )

        final_surface.blit(edges_and_corners['top_left'], (0, 0))
        final_surface.blit(edges_and_corners['top_right'], (width - corner_radius_param, 0))

        final_surface.blit(edges_and_corners['bottom_left'],
                           (0, height - corner_radius_param))
        final_surface.blit(edges_and_corners['bottom_right'],
                           (width - corner_radius_param, height - corner_radius_param))

        if width - (2 * corner_radius_param) > 0:
            top_edge = pygame.transform.scale(edges_and_corners['top'],
                                              (width - (2 * corner_radius_param),
                                               shadow_width_param))
            bottom_edge = pygame.transform.scale(edges_and_corners['bottom'],
                                                 (width - (2 * corner_radius_param),
                                                  shadow_width_param))
            final_surface.blit(top_edge, (corner_radius_param, 0))
            final_surface.blit(bottom_edge, (corner_radius_param, height - shadow_width_param))

        if height - (2 * corner_radius_param) > 0:
            left_edge = pygame.transform.scale(edges_and_corners['left'],
                                               (shadow_width_param,
                                                height - (2 * corner_radius_param)))
            right_edge = pygame.transform.scale(edges_and_corners['right'],
                                                (shadow_width_param,
                                                 height - (2 * corner_radius_param)))

            final_surface.blit(left_edge, (0, corner_radius_param))
            final_surface.blit(right_edge, (width - shadow_width_param,
                                            corner_radius_param))

        self._short_term_rect_cache[shadow_id] = final_surface
        return final_surface

    def create_new_ellipse_shadow(
        self,
        width: int,
        height: int,
        shadow_width_param: int,
        aa_amount: int = 4,
        color: Tuple3IntType = (0, 0, 0)
    ) -> Optional['pygame.Surface']:
        """
        Creates an ellipse shaped shadow surface at the specified size and stores it for later use.

        :param width: The width of the shadow to create
        :param height: The height of the shadow to create
        :param shadow_width_param: The width of the shadowed edge
        :param aa_amount: The amount of anti-aliasing to use, defaults to 4
        :param color: Shadow color (r, g, b)
        :return: Surface with shadow
        """
        assert isinstance(width, int)
        assert isinstance(height, int)
        assert_vector(color, 3, int)
        shadow_width_param, aa_amount = int(shadow_width_param), int(aa_amount)
        if shadow_width_param == 0:
            return None
        shadow_surface = pygame.surface.Surface((width * aa_amount, height * aa_amount),
                                                flags=pygame.SRCALPHA, depth=32)
        shadow_surface.fill(pygame.Color('#00000000'))
        r, g, b = color
        ellipse_id = str(width) + 'x' + str(height) + 'x' + str(shadow_width_param)
        if ellipse_id in self._created_ellipse_shadows:
            return self._created_ellipse_shadows[ellipse_id]

        alpha_increment = max(1, int(20 / shadow_width_param))
        shadow_alpha = alpha_increment
        shadow_width = width * aa_amount
        shadow_height = height * aa_amount
        for i in range(shadow_width_param):
            if shadow_width > 0 and shadow_height > 0:
                shadow_rect = pygame.Rect(i * aa_amount,
                                          i * aa_amount,
                                          shadow_width,
                                          shadow_height)
                pygame.draw.ellipse(shadow_surface,
                                    pygame.Color(r, g, b, shadow_alpha), shadow_rect)
                shadow_width -= (2 * aa_amount)
                shadow_height -= (2 * aa_amount)
                shadow_alpha += alpha_increment

        final_surface = pygame.transform.smoothscale(shadow_surface, (width, height))
        self._created_ellipse_shadows[ellipse_id] = final_surface
        return final_surface



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

CORE
This module contains the widget core classes.
"""

from pygame_menu.widgets.core.widget import Widget
from pygame_menu.widgets.core.selection import Selection



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

PYGAME-MENU HOOK
Used by Pyinstaller.
"""

import os

# noinspection PyProtectedMember
from pygame_menu import __file__ as pygame_menu_main_file

# Get pygame_menu's folder
pygame_menu_folder = os.path.dirname(os.path.abspath(pygame_menu_main_file))

# datas is the variable that pyinstaller looks for while processing hooks
datas = []


# A helper to append the relative path of a resource to hook variable - datas
def _append_to_datas(file_path: str, target_folder: str, base_target_folder: str = 'pygame_menu',
                     relative: bool = True) -> None:
    """
    Add path to datas.

    :param file_path: File path
    :param target_folder: Folder to paste the resources. If empty uses the containing folder of the file as ``base_target_folder+target_folder``
    :param base_target_folder: Base folder of the resource
    :param relative: If ``True`` append ``pygame_menu_folder``
    """
    global datas
    if relative:
        res_path = os.path.join(pygame_menu_folder, file_path)
    else:
        res_path = file_path
    if target_folder == '':
        target_folder = os.path.basename(os.path.dirname(res_path))
    if os.path.exists(res_path):
        datas.append((res_path, os.path.join(base_target_folder, target_folder)))


# Append data
from pygame_menu.font import FONT_EXAMPLES
from pygame_menu.baseimage import IMAGE_EXAMPLES
from pygame_menu.sound import SOUND_EXAMPLES

pygame_menu_resources = os.path.join('pygame_menu', 'resources')
for f in FONT_EXAMPLES:
    _append_to_datas(f, target_folder='', base_target_folder=pygame_menu_resources)
for f in IMAGE_EXAMPLES:
    _append_to_datas(f, target_folder='', base_target_folder=pygame_menu_resources)
for f in SOUND_EXAMPLES:
    _append_to_datas(f, target_folder='', base_target_folder=pygame_menu_resources)



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - MULTI-INPUT
Shows different inputs (widgets).
"""

__all__ = ['main']

import pygame
import pygame_menu
from pygame_menu.examples import create_example_window

from typing import Tuple, Optional

# Constants and global variables
FPS = 60
WINDOW_SIZE = (640, 480)

sound: Optional['pygame_menu.sound.Sound'] = None
surface: Optional['pygame.Surface'] = None
main_menu: Optional['pygame_menu.Menu'] = None


def main_background() -> None:
    """
    Background color of the main menu, on this function user can plot
    images, play sounds, etc.
    """
    surface.fill((40, 40, 40))


def check_name_test(value: str) -> None:
    """
    This function tests the text input widget.

    :param value: The widget value
    """
    print(f'User name: {value}')


def update_menu_sound(value: Tuple, enabled: bool) -> None:
    """
    Update menu sound.

    :param value: Value of the selector (Label and index)
    :param enabled: Parameter of the selector, (True/False)
    """
    assert isinstance(value, tuple)
    if enabled:
        main_menu.set_sound(sound, recursive=True)
        print('Menu sounds were enabled')
    else:
        main_menu.set_sound(None, recursive=True)
        print('Menu sounds were disabled')


def main(test: bool = False) -> None:
    """
    Main program.

    :param test: Indicate function is being tested
    """

    # -------------------------------------------------------------------------
    # Globals
    # -------------------------------------------------------------------------
    global main_menu
    global sound
    global surface

    # -------------------------------------------------------------------------
    # Create window
    # -------------------------------------------------------------------------
    surface = create_example_window('Example - Multi Input', WINDOW_SIZE)
    clock = pygame.time.Clock()

    # -------------------------------------------------------------------------
    # Set sounds
    # -------------------------------------------------------------------------
    sound = pygame_menu.sound.Sound()

    # Load example sounds
    sound.load_example_sounds()

    # Disable a sound
    sound.set_sound(pygame_menu.sound.SOUND_TYPE_ERROR, None)

    # -------------------------------------------------------------------------
    # Create menus: Settings
    # -------------------------------------------------------------------------
    settings_menu_theme = pygame_menu.themes.THEME_ORANGE.copy()
    settings_menu_theme.title_offset = (5, -2)
    settings_menu_theme.widget_alignment = pygame_menu.locals.ALIGN_LEFT
    settings_menu_theme.widget_font = pygame_menu.font.FONT_OPEN_SANS_LIGHT
    settings_menu_theme.widget_font_size = 20

    settings_menu = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.85,
        theme=settings_menu_theme,
        title='Settings',
        width=WINDOW_SIZE[0] * 0.9
    )

    # Add text inputs with different configurations
    settings_menu.add.text_input(
        'First name: ',
        default='John',
        onreturn=check_name_test,
        textinput_id='first_name'
    )
    settings_menu.add.text_input(
        'Last name: ',
        default='Rambo',
        maxchar=10,
        textinput_id='last_name',
        input_underline='.'
    )
    settings_menu.add.text_input(
        'Your age: ',
        default=25,
        maxchar=3,
        maxwidth=3,
        textinput_id='age',
        input_type=pygame_menu.locals.INPUT_INT,
        cursor_selection_enable=False
    )
    settings_menu.add.text_input(
        'Some long text: ',
        maxwidth=19,
        textinput_id='long_text',
        input_underline='_'
    )
    settings_menu.add.text_input(
        'Password: ',
        maxchar=6,
        password=True,
        textinput_id='pass',
        input_underline='_'
    )

    # Selectable items
    items = [('Easy', 'EASY'),
             ('Medium', 'MEDIUM'),
             ('Hard', 'HARD')]

    # Create selector with 3 difficulty options
    settings_menu.add.selector(
        'Select difficulty:\t',
        items,
        selector_id='difficulty',
        default=1
    )
    settings_menu.add.selector(
        'Select difficulty fancy',
        items,
        selector_id='difficulty_fancy',
        default=1,
        style='fancy'
    )
    settings_menu.add.dropselect(
        'Select difficulty (drop)',
        items,
        default=1,
        dropselect_id='difficulty_drop'
    )
    settings_menu.add.dropselect_multiple(
        title='Pick 3 colors',
        items=[('Black', (0, 0, 0)),
               ('Blue', (0, 0, 255)),
               ('Cyan', (0, 255, 255)),
               ('Fuchsia', (255, 0, 255)),
               ('Green', (0, 255, 0)),
               ('Red', (255, 0, 0)),
               ('White', (255, 255, 255)),
               ('Yellow', (255, 255, 0))],
        dropselect_multiple_id='pickcolors',
        max_selected=3,
        open_middle=True,
        selection_box_height=6  # How many options show if opened
    )

    # Create switch
    settings_menu.add.toggle_switch('First Switch', False,
                                    toggleswitch_id='first_switch')
    settings_menu.add.toggle_switch('Other Switch', True,
                                    toggleswitch_id='second_switch',
                                    state_text=('Apagado', 'Encendido'))

    # Single value from range
    rslider = settings_menu.add.range_slider('Choose a number', 50, (0, 100), 1,
                                             rangeslider_id='range_slider',
                                             value_format=lambda x: str(int(x)))

    # Range
    settings_menu.add.range_slider('How would you rate pygame-menu?', (7, 10), (1, 10), 1,
                                   rangeslider_id='range_slider_double')

    # Create discrete range
    range_values_discrete = {0: 'A', 1: 'B', 2: 'C', 3: 'D', 4: 'E', 5: 'F'}
    settings_menu.add.range_slider('Pick a letter', 0, list(range_values_discrete.keys()),
                                   rangeslider_id='range_slider_discrete',
                                   slider_text_value_enabled=False,
                                   value_format=lambda x: range_values_discrete[x])

    # Add a progress bar
    progress = settings_menu.add.progress_bar('Progress', default=rslider.get_value(), progressbar_id='progress')

    def on_change_slider(val: int) -> None:
        """
        Updates the progress bar.

        :param val: Value of the progress from 0 to 100
        """
        progress.set_value(val)

    rslider.set_onchange(on_change_slider)

    # Add a block
    settings_menu.add.clock(clock_format='%Y/%m/%d %H:%M', title_format='Clock: {0}')

    def data_fun() -> None:
        """
        Print data of the menu.
        """
        print('Settings data:')
        data = settings_menu.get_input_data()
        for k in data.keys():
            print(f'\t{k}\t=>\t{data[k]}')

    # Add final buttons
    settings_menu.add.button('Store data', data_fun, button_id='store')  # Call function
    settings_menu.add.button('Restore original values', settings_menu.reset_value)
    settings_menu.add.button('Return to main menu', pygame_menu.events.BACK,
                             align=pygame_menu.locals.ALIGN_CENTER)

    # -------------------------------------------------------------------------
    # Create menus: More settings
    # -------------------------------------------------------------------------
    more_settings_menu = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.85,
        theme=settings_menu_theme,
        title='More Settings',
        width=WINDOW_SIZE[0] * 0.9
    )

    more_settings_menu.add.image(
        pygame_menu.baseimage.IMAGE_EXAMPLE_PYGAME_MENU,
        scale=(0.25, 0.25),
        align=pygame_menu.locals.ALIGN_CENTER
    )
    more_settings_menu.add.color_input(
        'Color 1 RGB: ',
        color_type='rgb'
    )
    more_settings_menu.add.color_input(
        'Color 2 RGB: ',
        color_type='rgb',
        default=(255, 0, 0),
        input_separator='-'
    )

    def print_color(color: Tuple) -> None:
        """
        Test onchange/onreturn.

        :param color: Color tuple
        """
        print('Returned color: ', color)

    more_settings_menu.add.color_input(
        'Color in Hex: ',
        color_type='hex',
        hex_format='lower',
        color_id='hex_color',
        onreturn=print_color
    )

    more_settings_menu.add.vertical_margin(25)
    more_settings_menu.add.button(
        'Return to main menu',
        pygame_menu.events.BACK,
        align=pygame_menu.locals.ALIGN_CENTER
    )

    # -------------------------------------------------------------------------
    # Create menus: Column buttons
    # -------------------------------------------------------------------------
    button_column_menu_theme = pygame_menu.themes.THEME_ORANGE.copy()
    button_column_menu_theme.background_color = pygame_menu.BaseImage(
        image_path=pygame_menu.baseimage.IMAGE_EXAMPLE_GRAY_LINES,
        drawing_mode=pygame_menu.baseimage.IMAGE_MODE_REPEAT_XY
    )
    button_column_menu_theme.widget_font_size = 25

    button_column_menu = pygame_menu.Menu(
        columns=2,
        height=WINDOW_SIZE[1] * 0.45,
        rows=3,
        theme=button_column_menu_theme,
        title='Textures+Columns',
        width=WINDOW_SIZE[0] * 0.9
    )
    for i in range(4):
        button_column_menu.add.button(f'Button {i}', pygame_menu.events.BACK)
    button_column_menu.add.button(
        'Return to main menu', pygame_menu.events.BACK,
        background_color=pygame_menu.BaseImage(
            image_path=pygame_menu.baseimage.IMAGE_EXAMPLE_METAL
        )
    ).background_inflate_to_selection_effect()

    # -------------------------------------------------------------------------
    # Create menus: Main menu
    # -------------------------------------------------------------------------
    main_menu_theme = pygame_menu.themes.THEME_ORANGE.copy()
    main_menu_theme.title_font = pygame_menu.font.FONT_COMIC_NEUE
    main_menu_theme.widget_font = pygame_menu.font.FONT_COMIC_NEUE
    main_menu_theme.widget_font_size = 30

    main_menu = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.7,
        onclose=pygame_menu.events.EXIT,  # User press ESC button
        theme=main_menu_theme,
        title='Main menu',
        width=WINDOW_SIZE[0] * 0.8
    )

    main_menu.add.button('Settings', settings_menu)
    main_menu.add.button('More Settings', more_settings_menu)
    main_menu.add.button('Menu in textures and columns', button_column_menu)
    main_menu.add.selector('Menu sounds ',
                           [('Off', False), ('On', True)],
                           onchange=update_menu_sound)
    main_menu.add.button('Quit', pygame_menu.events.EXIT)

    # -------------------------------------------------------------------------
    # Main loop
    # -------------------------------------------------------------------------
    while True:

        # Tick
        clock.tick(FPS)

        # Paint background
        main_background()

        # Main menu
        main_menu.mainloop(surface, main_background, disable_loop=test, fps_limit=FPS)

        # Flip surface
        pygame.display.flip()

        # At first loop returns
        if test:
            break


if __name__ == '__main__':
    main()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

WIDGETS
Widgets elements that can be added to the Menu.
"""

# Widgets core
import pygame_menu.widgets.core
from pygame_menu.widgets.core import Widget

# Selection
from pygame_menu.widgets.selection import HighlightSelection, LeftArrowSelection, \
    NoneSelection, RightArrowSelection, SimpleSelection

# Widgets
from pygame_menu.widgets.widget import Button, ColorInput, DropSelect, \
    DropSelectMultiple, Frame, HMargin, Image, Label, MenuBar, MenuLink, NoneWidget, \
    ProgressBar, RangeSlider, ScrollBar, Selector, SurfaceWidget, Table, TextInput, \
    ToggleSwitch, VFill, VMargin

# Widget constants
from pygame_menu.widgets.widget.colorinput import COLORINPUT_TYPE_RGB, \
    COLORINPUT_TYPE_HEX, COLORINPUT_HEX_FORMAT_UPPER, COLORINPUT_HEX_FORMAT_NONE, \
    COLORINPUT_HEX_FORMAT_LOWER

from pygame_menu.widgets.widget.dropselect_multiple import \
    DROPSELECT_MULTIPLE_SFORMAT_TOTAL, DROPSELECT_MULTIPLE_SFORMAT_LIST_COMMA, \
    DROPSELECT_MULTIPLE_SFORMAT_LIST_HYPHEN

from pygame_menu.widgets.widget.frame import FRAME_DEFAULT_TITLE_BACKGROUND_COLOR, \
    FRAME_TITLE_BUTTON_CLOSE, FRAME_TITLE_BUTTON_MAXIMIZE, FRAME_TITLE_BUTTON_MINIMIZE

from pygame_menu.widgets.widget.menubar import MENUBAR_STYLE_ADAPTIVE, \
    MENUBAR_STYLE_SIMPLE, MENUBAR_STYLE_TITLE_ONLY, MENUBAR_STYLE_TITLE_ONLY_DIAGONAL, \
    MENUBAR_STYLE_NONE, MENUBAR_STYLE_UNDERLINE, MENUBAR_STYLE_UNDERLINE_TITLE

from pygame_menu.widgets.widget.selector import SELECTOR_STYLE_CLASSIC, \
    SELECTOR_STYLE_FANCY



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

HORIZONTAL MARGIN
Horizontal box margin.
"""

__all__ = [
    'HMargin',
    'HMarginManager'
]

import pygame
import pygame_menu

from abc import ABC
from pygame_menu.widgets.core.widget import AbstractWidgetManager
from pygame_menu.widgets.widget.none import NoneWidget

from pygame_menu._types import NumberType, NumberInstance


# noinspection PyMissingOrEmptyDocstring
class HMargin(NoneWidget):
    """
    Horizontal margin widget.

    .. note::

        HMargin does not accept any transformation.

    :param margin: Horizontal margin in px
    :param widget_id: ID of the widget
    """

    def __init__(
        self,
        margin: NumberType,
        widget_id: str = ''
    ) -> None:
        assert isinstance(margin, NumberInstance)
        assert margin > 0, \
            'zero margin is not valid, prefer adding a NoneWidget menu.add.none_widget()'
        super(HMargin, self).__init__(widget_id=widget_id)
        self._rect.width = int(margin)
        self._rect.height = 0

    def get_rect(self, *args, **kwargs) -> 'pygame.Rect':
        return self._rect.copy()


class HMarginManager(AbstractWidgetManager, ABC):
    """
    HMargin manager.
    """

    def horizontal_margin(
        self,
        margin: NumberType,
        margin_id: str = ''
    ) -> 'pygame_menu.widgets.HMargin':
        """
        Adds a horizontal margin to the Menu. Only useful in frames.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param margin: Horizontal margin in px
        :param margin_id: ID of the horizontal margin
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.HMargin`
        """
        attributes = self._filter_widget_attributes({})
        widget = HMargin(margin, widget_id=margin_id)
        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)

        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - DYNAMIC WIDGET UPDATE
Dynamically updates the widgets based on user events.
"""

__all__ = ['main']

import pygame
import pygame_menu
from pygame_menu.examples import create_example_window

import math
from typing import Dict, Any


class App(object):
    """
    The following object creates the whole app.
    """
    image_widget: 'pygame_menu.widgets.Image'
    item_description_widget: 'pygame_menu.widgets.Label'
    menu: 'pygame_menu.Menu'
    modes: Dict[int, Dict[str, Any]]
    quit_button: 'pygame_menu.widgets.Button'
    quit_button_fake: 'pygame_menu.widgets.Button'
    selector_widget: 'pygame_menu.widgets.Selector'
    surface: 'pygame.Surface'

    def __init__(self) -> None:
        self.surface = create_example_window('Example - Dynamic Widget Update', (640, 480), flags=pygame.NOFRAME)

        # Load image
        default_image = pygame_menu.BaseImage(
            image_path=pygame_menu.baseimage.IMAGE_EXAMPLE_PYGAME_MENU
        ).scale(0.2, 0.2)

        # Set theme
        theme = pygame_menu.themes.THEME_DEFAULT.copy()
        theme.title_bar_style = pygame_menu.widgets.MENUBAR_STYLE_UNDERLINE_TITLE
        theme.title_close_button_cursor = pygame_menu.locals.CURSOR_HAND
        theme.title_font_color = (35, 35, 35)

        # This dict stores the values of the widgets to be changed dynamically
        self.modes = {
            1: {
                'image': default_image.copy(),
                'label': {
                    'color': theme.widget_font_color,
                    'size': theme.widget_font_size,
                    'text': 'The first one is very epic'
                }
            },
            2: {
                'image': default_image.copy().to_bw(),
                'label': {
                    'color': (0, 0, 0),
                    'size': 20,
                    'text': 'This other one is also epic, but fancy'
                }
            },
            3: {
                'image': default_image.copy().flip(False, True).pick_channels('r'),
                'label': {
                    'color': (255, 0, 0),
                    'size': 45,
                    'text': 'YOU D I E D'
                }
            }
        }

        # Create menus
        self.menu = pygame_menu.Menu(
            height=480,
            onclose=pygame_menu.events.CLOSE,
            theme=theme,
            title='Everything is dynamic now',
            width=640
        )

        self.selector_widget = self.menu.add.selector(
            title='Pick one option: ',
            items=[('The first', 1),
                   ('The second', 2),
                   ('The final mode', 3)],
            onchange=self._on_selector_change
        )

        self.image_widget = self.menu.add.image(
            image_path=self.modes[1]['image'],
            padding=(25, 0, 0, 0)  # top, right, bottom, left
        )

        self.item_description_widget = self.menu.add.label(title='')

        self.quit_button = self.menu.add.button('Quit', pygame_menu.events.EXIT)

        self.quit_button_fake = self.menu.add.button('You cannot quit', self.fake_quit, font_color=(255, 255, 255))
        self.quit_button_fake.add_draw_callback(self.animate_quit_button)

        # Update the widgets based on selected value from selector get_value
        # returns selected item tuple and index, so [0][1] means the second object
        # from ('The first', 1) tuple
        self._update_from_selection(int(self.selector_widget.get_value()[0][1]))

    def animate_quit_button(
        self,
        widget: 'pygame_menu.widgets.Widget',
        menu: 'pygame_menu.Menu'
    ) -> None:
        """
        Animate widgets if the last option is selected.

        :param widget: Widget to be updated
        :param menu: Menu
        """
        if self.current == 3:
            t = widget.get_counter_attribute('t', menu.get_clock().get_time() * 0.0075, math.pi)
            widget.set_padding(10 * (1 + math.sin(t)))  # Oscillating padding
            widget.set_background_color((int(125 * (1 + math.sin(t))), 0, 0), None)
            c = int(127 * (1 + math.cos(t)))
            widget.update_font({'color': (c, c, c)})  # Widget font now is in grayscale
            # widget.translate(10 * math.cos(t), 10 * math.sin(t))
            widget.rotate(5 * t)

    @staticmethod
    def fake_quit() -> None:
        """
        Function executed by fake quit button.
        """
        print('I said that you cannot quit')

    def _update_from_selection(self, index: int) -> None:
        """
        Change widgets depending on index.

        :param index: Index
        """
        self.current = index
        self.image_widget.set_image(self.modes[index]['image'])
        self.item_description_widget.set_title(self.modes[index]['label']['text'])
        self.item_description_widget.update_font(
            {'color': self.modes[index]['label']['color'],
             'size': self.modes[index]['label']['size']}
        )
        # Swap buttons using hide/show
        if index == 3:
            self.quit_button.hide()
            self.quit_button_fake.show()
        else:
            self.quit_button.show()
            self.quit_button_fake.hide()

    def _on_selector_change(self, selected: Any, value: int) -> None:
        """
        Function executed if selector changes.

        :param selected: Selector data containing text and index
        :param value: Value from the selected option
        """
        print('Selected data:', selected)
        self._update_from_selection(value)

    def mainloop(self, test: bool) -> None:
        """
        App mainloop.

        :param test: Test status
        """
        self.menu.mainloop(self.surface, disable_loop=test)


def main(test: bool = False) -> 'App':
    """
    Main function.

    :param test: Indicate function is being tested
    :return: App object
    """
    app = App()
    app.mainloop(test)
    return app


if __name__ == '__main__':
    main()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

COLOR INPUT
Color input class, Widget created in top of TextInput that provides a textbox
for entering and previewing colors in RGB and HEX format.
"""

__all__ = [

    # Main class
    'ColorInput',
    'ColorInputManager',

    # Constants
    'COLORINPUT_TYPE_HEX',
    'COLORINPUT_TYPE_RGB',
    'COLORINPUT_HEX_FORMAT_LOWER',
    'COLORINPUT_HEX_FORMAT_NONE',
    'COLORINPUT_HEX_FORMAT_UPPER',

    # Type
    'ColorInputColorType',
    'ColorInputHexFormatType'

]

import math
import pygame
import pygame_menu

from abc import ABC
from pygame_menu.locals import INPUT_TEXT
from pygame_menu.utils import check_key_pressed_valid, make_surface
from pygame_menu.widgets.core.widget import AbstractWidgetManager, Widget
from pygame_menu.widgets.widget.textinput import TextInput

from pygame_menu._types import Union, List, NumberType, Any, Optional, CallbackType, \
    Tuple3IntType, NumberInstance, EventVectorType, Callable

# Input modes
COLORINPUT_TYPE_HEX = 'hex'
COLORINPUT_TYPE_RGB = 'rgb'

# Apply format to hex color string
COLORINPUT_HEX_FORMAT_LOWER = 'lower'
COLORINPUT_HEX_FORMAT_NONE = 'none'
COLORINPUT_HEX_FORMAT_UPPER = 'upper'

# Custom typing
ColorInputColorType = str
ColorInputHexFormatType = str


# noinspection PyMissingOrEmptyDocstring
class ColorInput(TextInput):
    """
    Color input widget.

    The callbacks receive the current value and all unknown keyword arguments,
    where ``current_color=widget.get_value()``:

    .. code-block:: python

        onchange(current_color, **kwargs)
        onreturn(current_color, **kwargs)

    .. note::

        This widget cannot select text as :py:class:`pygame_menu.widgets.TextInput`
        does. Also, copy and paste is disabled.

    .. note::

        ColorInput accepts the same transformations as :py:class:`pygame_menu.widgets.TextInput`.

    :param title: Color input title
    :param colorinput_id: ID of the text input
    :param color_type: Type of color input
    :param cursor_switch_ms: Interval of cursor switch between off and on status. First status is ``off``
    :param dynamic_width: If ``True`` the widget width changes if the previsualization color box is active or not
    :param hex_format: Hex format string mode
    :param input_separator: Divisor between RGB channels
    :param input_underline: Character drawn under each number input
    :param input_underline_vmargin: Vertical margin of underline in px
    :param cursor_color: Color of cursor
    :param onchange: Function when changing the values of the color text
    :param onreturn: Function when pressing return (apply) on the color text input
    :param onselect: Function when selecting the widget
    :param prev_margin: Horizontal margin between the previsualization and the input text in px
    :param prev_width_factor: Width of the previsualization box in terms of the height of the widget
    :param repeat_keys_initial_ms: Time in milliseconds before keys are repeated when held
    :param repeat_keys_interval_ms: Interval between key press repetition when held
    :param repeat_mouse_interval_ms: Interval between mouse events when held
    :param kwargs: Optional keyword arguments
    """
    _auto_separator_pos: List[int]
    _color_type: str
    _dynamic_width: bool
    _hex_format: str
    _last_g: int
    _last_r: int
    _prev_margin: int
    _previsualization_surface: Optional['pygame.Surface']
    _separator: str

    def __init__(
        self,
        title: Any,
        colorinput_id: str = '',
        color_type: ColorInputColorType = COLORINPUT_TYPE_RGB,
        cursor_color: Tuple3IntType = (0, 0, 0),
        cursor_switch_ms: NumberType = 500,
        dynamic_width: bool = True,
        hex_format: ColorInputHexFormatType = COLORINPUT_HEX_FORMAT_NONE,
        input_separator: str = ',',
        input_underline: str = '_',
        input_underline_vmargin: int = 0,
        onchange: CallbackType = None,
        onreturn: CallbackType = None,
        onselect: CallbackType = None,
        prev_margin: int = 10,
        prev_width_factor: NumberType = 3,
        repeat_keys_initial_ms: NumberType = 450,
        repeat_keys_interval_ms: NumberType = 80,
        repeat_mouse_interval_ms: NumberType = 100,
        *args,
        **kwargs
    ) -> None:
        assert isinstance(color_type, str)
        assert isinstance(colorinput_id, str)
        assert isinstance(dynamic_width, bool)
        assert isinstance(hex_format, str)
        assert isinstance(input_separator, str)
        assert isinstance(input_underline, str)
        assert isinstance(prev_margin, int)
        assert isinstance(prev_width_factor, NumberInstance)

        assert len(input_separator) == 1, 'input_separator must be a single char'
        assert len(input_separator) != 0, 'input_separator cannot be empty'
        assert prev_width_factor > 0, \
            'previsualization width factor must be greater than zero'
        assert input_separator not in ('0', '1', '2', '3', '4', '5', '6', '7', '8',
                                       '9'), 'input_separator cannot be a number'
        assert color_type in (COLORINPUT_TYPE_HEX, COLORINPUT_TYPE_RGB), \
            f'color type must be "{COLORINPUT_TYPE_HEX}" or "{COLORINPUT_TYPE_RGB}"'
        assert hex_format in (COLORINPUT_HEX_FORMAT_NONE, COLORINPUT_HEX_FORMAT_LOWER,
                              COLORINPUT_HEX_FORMAT_UPPER), \
            'invalid hex format mode, it must be "none", "lower" or "upper"'

        _maxchar = 0
        self._color_type = color_type.lower()
        if self._color_type == COLORINPUT_TYPE_RGB:
            _maxchar = 11  # RRR,GGG,BBB
            self._valid_chars = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                                 input_separator]
        elif self._color_type == COLORINPUT_TYPE_HEX:
            _maxchar = 7  # #XXYYZZ
            self._valid_chars = ['a', 'A', 'b', 'B', 'c', 'C', 'd', 'D', 'e', 'E',
                                 'f', 'F', '#', '0', '1', '2', '3', '4', '5', '6',
                                 '7', '8', '9']

        # noinspection PyArgumentEqualDefault
        super(ColorInput, self).__init__(
            copy_paste_enable=False,
            cursor_color=cursor_color,
            cursor_switch_ms=cursor_switch_ms,
            cursor_selection_enable=False,
            history=0,
            input_type=INPUT_TEXT,
            input_underline=input_underline,
            input_underline_vmargin=input_underline_vmargin,
            maxchar=_maxchar,
            maxwidth=0,
            onchange=onchange,
            onreturn=onreturn,
            onselect=onselect,
            password=False,
            repeat_keys_initial_ms=repeat_keys_initial_ms,
            repeat_keys_interval_ms=repeat_keys_interval_ms,
            repeat_mouse_interval_ms=repeat_mouse_interval_ms,
            text_ellipsis='',
            textinput_id=colorinput_id,
            title=title,
            valid_chars=self._valid_chars,
            *args,
            **kwargs
        )

        # Store inner variables
        self._auto_separator_pos = []  # This stores indexes of auto separator added
        self._dynamic_width = dynamic_width
        self._hex_format = hex_format
        self._separator = input_separator

        # Previsualization surface, if -1 does not show
        self._last_b = -1
        self._last_g = -1
        self._last_r = -1
        self._prev_margin = prev_margin
        self._prev_width_factor = prev_width_factor
        self._previsualization_surface = None

        # Disable parent callbacks
        self._apply_widget_update_callback = False

        # Disable alt+x
        self._alt_x_enabled = False

    def _apply_font(self) -> None:
        super(ColorInput, self)._apply_font()

        # Compute the size of the underline
        if self._input_underline != '':
            max_width = 0  # Max expected width
            if self._color_type == COLORINPUT_TYPE_RGB:
                max_width = self._font_render_string(
                    f'255{self._separator}255{self._separator}255'
                ).get_width()
            else:
                for i in ('a', 'b', 'c', 'd', 'e', 'f'):
                    max_width = max(
                        max_width,
                        self._font_render_string(f'#{i * 6}').get_width()
                    )

            char = math.ceil(max_width / self._input_underline_size)
            for i in range(10):  # Find the best guess for
                fw = self._font_render_string(self._input_underline * int(char)).get_width()
                char += 1
                if fw >= max_width:
                    break

            self._input_underline_len = char

    def clear(self) -> None:
        super(ColorInput, self).clear()
        self._previsualization_surface = None
        self._auto_separator_pos = []
        if self._color_type == COLORINPUT_TYPE_HEX:
            super(ColorInput, self).set_value('#')
        self.change()

    def set_value(self, color: Optional[Union[str, Tuple3IntType]]) -> None:
        """
        Set the color value.

        :param color: A string if the type is HEX, or a (r, g, b) tuple if RGB
        """
        if color is None:
            color = ''
        format_color = ''
        if self._color_type == COLORINPUT_TYPE_RGB:
            if color == '':
                super(ColorInput, self).set_value('')
                return
            assert isinstance(color, tuple), \
                'color in rgb format must be a tuple in (r,g,b) format'
            assert len(color) == 3, 'tuple must contain only 3 colors, R,G,B'
            r, g, b = color
            assert isinstance(r, int), 'red color must be an integer'
            assert isinstance(g, int), 'blue color must be an integer'
            assert isinstance(b, int), 'green color must be an integer'
            assert 0 <= r <= 255, 'red color must be between 0 and 255'
            assert 0 <= g <= 255, 'blue color must be between 0 and 255'
            assert 0 <= b <= 255, 'green color must be between 0 and 255'
            format_color = f'{r}{self._separator}{g}{self._separator}{b}'
            self._auto_separator_pos = [0, 1]

        elif self._color_type == COLORINPUT_TYPE_HEX:
            text = str(color).strip()
            if text == '' or text == '#':
                format_color = '#'
            else:
                # Remove all invalid chars
                valid_text = ''
                for ch in text:
                    if ch in self._valid_chars:
                        valid_text += ch
                text = valid_text

                # Check if the color is valid
                count_hash = 0
                for ch in text:
                    if ch == '#':
                        count_hash += 1
                if count_hash == 1:
                    assert text[0] == '#', 'color format must be "#RRGGBB"'
                if count_hash == 0:
                    text = '#' + text
                assert len(text) == 7, \
                    'invalid color, only formats "#RRGGBB" or "RRGGBB" are allowed'
                format_color = text

        super(ColorInput, self).set_value(format_color)
        self._format_hex()

    def value_changed(self) -> bool:
        default = self._default_value
        if self._color_type == COLORINPUT_TYPE_HEX and '#' not in default:
            default = '#' + default
        return self.get_value(as_string=True) != default

    def get_value(self, as_string: bool = False) -> Union[str, Tuple3IntType]:
        """
        Return the color value as a tuple or red blue and green channels.

        .. note::

            If the data is invalid the widget returns ``(-1, -1, -1)``.

        :param as_string: If ``True`` returns the widget value as plain text
        :return: Color tuple as (R, G, B) or color string
        """
        assert isinstance(as_string, bool)
        if as_string:
            return self._input_string

        if self._color_type == COLORINPUT_TYPE_RGB:
            color = self._input_string.split(self._separator)
            if len(color) == 3 and color[0] != '' and color[1] != '' and color[2] != '':
                r, g, b = int(color[0]), int(color[1]), int(color[2])
                if 0 <= r <= 255 and 0 <= g <= 255 and 0 <= g <= 255:
                    return r, g, b

        elif self._color_type == COLORINPUT_TYPE_HEX:
            if len(self._input_string) == 7:
                color = self._input_string[1:]
                color = tuple(int(color[i:i + 2], 16) for i in (0, 2, 4))
                return color[0], color[1], color[2]

        return -1, -1, -1

    def is_valid(self) -> bool:
        """
        Return ``True`` if the current value of the input is a valid color or not.

        :return: ``True`` if valid
        """
        r, g, b = self.get_value()
        if r == -1 or g == -1 or b == -1:
            return False
        return True

    def _draw(self, surface: 'pygame.Surface') -> None:
        super(ColorInput, self)._draw(surface)  # This calls _render()

        # Draw previsualization box
        if self._previsualization_surface is not None:
            posx = self._rect.x + self._rect.width \
                   - self._prev_width_factor * self._rect.height \
                   + self._rect.height / 10
            posy = self._rect.y
            surface.blit(self._previsualization_surface, (int(posx), int(posy)))

    def _render(self) -> Optional[bool]:
        render_text = super(ColorInput, self)._render()

        # Maybe TextInput did not render, so this has to be changed
        self._rect.width, self._rect.height = self._surface.get_size()
        if not self._dynamic_width or (self._dynamic_width and self._previsualization_surface is not None):
            self._rect.width += self._prev_width_factor * self._rect.height + self._prev_margin

        # Render the previsualization box
        r, g, b = self.get_value()
        if not self.is_valid():  # Remove previsualization if invalid color
            self._previsualization_surface = None
            return render_text

        # If previsualization surface is None or the color changed
        if self._last_r != r or self._last_b != b or self._last_g != g or self._previsualization_surface is None:
            width = self._prev_width_factor * self._rect.height
            if width == 0 or self._rect.height == 0:
                self._previsualization_surface = None
            else:
                self._previsualization_surface = make_surface(width, self._rect.height)
                self._previsualization_surface.fill((r, g, b))
                self._last_r = r
                self._last_g = g
                self._last_b = b
                if self._dynamic_width:
                    self._rect.width += self._prev_width_factor * self._rect.height + self._prev_margin

        return render_text

    def _format_hex(self) -> None:
        """
        Apply hex format.
        """
        if self._color_type != COLORINPUT_TYPE_HEX or self._hex_format == COLORINPUT_HEX_FORMAT_NONE:
            return
        elif self._hex_format == COLORINPUT_HEX_FORMAT_LOWER:
            self._input_string = self._input_string.lower()
        elif self._hex_format == COLORINPUT_HEX_FORMAT_UPPER:
            self._input_string = self._input_string.upper()

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)

        if self.readonly or not self.is_visible():
            self._readonly_check_mouseover(events)
            return False

        input_str = self._input_string
        cursor_pos = self._cursor_position
        disable_remove_separator = True

        key = ''  # Pressed key

        if self._color_type == COLORINPUT_TYPE_RGB:
            for event in events:

                # User writes
                if event.type == pygame.KEYDOWN and self._keyboard_enabled:
                    # Check if any key is pressed
                    if self._ignores_keyboard_nonphysical() and not check_key_pressed_valid(event):
                        continue

                    if disable_remove_separator and len(input_str) > 0 and len(input_str) > cursor_pos and (
                        f'{self._separator}{self._separator}' not in input_str or
                        input_str[cursor_pos] == self._separator and len(input_str) == cursor_pos + 1
                    ):
                        # Backspace button, delete text from right
                        if self._ctrl.back(event, self):
                            if len(input_str) >= 1 and input_str[cursor_pos - 1] == self._separator:
                                return True

                        # Delete button, delete text from left
                        elif self._ctrl.delete(event, self):
                            if input_str[cursor_pos] == self._separator:
                                return True

                    # Verify only on user key input, the rest of events are checked
                    # by TextInput on super call
                    key = str(event.unicode)
                    if key in self._valid_chars:
                        new_string = (
                            self._input_string[:self._cursor_position]
                            + key
                            + self._input_string[self._cursor_position:]
                        )

                        # Cannot be separator at first
                        if len(input_str) == 0 and key == self._separator:
                            return False

                        if len(input_str) > 1:
                            # Check separators
                            if key == self._separator:
                                # If more than 2 separators
                                total_separator = 0
                                for ch in input_str:
                                    if ch == self._separator:
                                        total_separator += 1
                                if total_separator >= 2:
                                    return False

                            # Check the number between the current separators,
                            # this number must be between 0-255
                            if key != self._separator:
                                pos_before = 0
                                pos_after = 0
                                for i in range(cursor_pos):
                                    if new_string[cursor_pos - i - 1] == self._separator:
                                        pos_before = cursor_pos - i
                                        break
                                for i in range(len(new_string) - cursor_pos):
                                    if new_string[cursor_pos + i] == self._separator:
                                        pos_after = cursor_pos + i
                                        break
                                if pos_after == 0:
                                    pos_after = len(new_string)
                                num = new_string[pos_before:pos_after].replace(',', '')
                                if num == '':
                                    num = '0'

                                if int(num) > 255:  # Number exceeds 25X
                                    return False
                                # User adds 0 at left, example: 12 -> 012
                                if num != str(int(num)) and key == '0':
                                    return False
                                if len(num) > 3:  # Number like 0XXX
                                    return False

        elif self._color_type == COLORINPUT_TYPE_HEX:
            self._format_hex()

            for event in events:
                # User writes
                if event.type == pygame.KEYDOWN and self._keyboard_enabled:
                    # Check if any key is pressed
                    if self._ignores_keyboard_nonphysical() and not check_key_pressed_valid(event):
                        continue

                    # Backspace button, delete text from right
                    if self._ctrl.back(event, self):
                        if cursor_pos == 1:
                            return True

                    # Delete button, delete text from left
                    elif self._ctrl.delete(event, self):
                        if cursor_pos == 0:
                            return True

                    # Verify only on user key input, the rest of events are checked
                    # by TextInput on super call
                    key = str(event.unicode)
                    if key in self._valid_chars:
                        if key == '#':
                            return True
                        if cursor_pos == 0:
                            return True

        # Update
        updated = super(ColorInput, self).update(events)

        # After
        if self._color_type == COLORINPUT_TYPE_RGB:
            total_separator = 0
            for ch in input_str:
                if ch == self._separator:
                    total_separator += 1

            # Adds auto separator
            if key == '0' and len(self._input_string) == self._cursor_position and total_separator < 2 and (
                len(self._input_string) == 1 or
                len(self._input_string) > 2 and self._input_string[self._cursor_position - 2] == self._separator
            ):
                self._push_key_input(self._separator, sounds=False)  # This calls .onchange()

            # Check number is valid (fix) because sometimes the user can type
            # too fast and avoid analysis of the text
            colors = self._input_string.split(self._separator)
            for c in colors:
                if len(c) > 0 and (int(c) > 255 or int(c) < 0):
                    self._input_string = input_str
                    self._cursor_position = cursor_pos
                    break

            if len(colors) == 3:
                self._auto_separator_pos = [0, 1]

            # Add an auto separator if the number can't continue growing and the cursor
            # is at the end of the line
            if total_separator < 2 and len(self._input_string) == self._cursor_position:
                auto_pos = len(colors) - 1
                last_num = colors[auto_pos]
                if (
                    ((len(last_num) == 2 and int(last_num) > 25) or (len(last_num) == 3 and int(last_num) <= 255)) and
                    auto_pos not in self._auto_separator_pos
                ):
                    self._push_key_input(self._separator, sounds=False)  # This calls .onchange()
                    self._auto_separator_pos.append(auto_pos)

            # If the user cleared all the string, reset auto separator
            if total_separator == 0 and (
                len(self._input_string) < 2 or
                len(self._input_string) == 2 and int(colors[0]) <= 25
            ):
                self._auto_separator_pos = []

        return updated


class ColorInputManager(AbstractWidgetManager, ABC):
    """
    ColorInput manager.
    """

    def color_input(
        self,
        title: Union[str, Any],
        color_type: ColorInputColorType,
        color_id: str = '',
        default: Union[str, Tuple3IntType] = '',
        hex_format: ColorInputHexFormatType = COLORINPUT_HEX_FORMAT_NONE,
        input_separator: str = ',',
        input_underline: str = '_',
        onchange: CallbackType = None,
        onreturn: CallbackType = None,
        onselect: Optional[Callable[[bool, 'Widget', 'pygame_menu.Menu'], Any]] = None,
        **kwargs
    ) -> 'pygame_menu.widgets.ColorInput':
        """
        Add a color widget with RGB or HEX format to the Menu.
        Includes a preview box that renders the given color.

        The callbacks (if defined) receive the current value and all unknown
        keyword arguments, where ``current_color=widget.get_value()``:

        .. code-block:: python

            onchange(current_color, **kwargs)
            onreturn(current_color, **kwargs)

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``dynamic_width``                 (bool) – If ``True`` the widget width changes if the pre-visualization color box is active or not
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``input_underline_vmargin``       (int) – Vertical margin of underline in px
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``maxwidth_dynamically_update``   (bool) - Dynamically update maxwidth depending on char size. ``True`` by default
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``previsualization_margin``       (int) – Pre-visualization left margin from text input in px. Default is ``0``
            - ``previsualization_width``        (int, float) – Pre-visualization width as a factor of the height. Default is ``3``
            - ``readonly_color``                (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode
            - ``readonly_selected_color``       (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode and is selected
            - ``repeat_keys``                   (bool) - Enable key repeat. ``True`` by default
            - ``repeat_keys_initial_ms``        (int, float) - Time in milliseconds before keys are repeated when held in milliseconds. ``400`` by default
            - ``repeat_keys_interval_ms``       (int, float) - Interval between key press repetition when held in milliseconds. ``50`` by default
            - ``repeat_mouse_interval_ms``      (int, float) - Interval between mouse events when held in milliseconds. ``400`` by default
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled
            - ``tab_size``                      (int) – Width of a tab character

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        .. warning::

            Be careful with kwargs collision. Consider that all optional documented
            kwargs keys are removed from the object.

        :param title: Title of the color input
        :param color_type: Type of the color input
        :param color_id: ID of the color input
        :param default: Default value to display, if RGB type it must be a tuple ``(r, g, b)``, if HEX must be a string ``"#XXXXXX"``
        :param hex_format: Hex format string mode
        :param input_separator: Divisor between RGB channels, not valid in HEX format
        :param input_underline: Underline character
        :param onchange: Callback executed when changing the values of the color text
        :param onreturn: Callback executed when pressing return (apply) on the color input
        :param onselect: Callback executed when selecting the widget
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.ColorInput`
        """
        assert isinstance(default, (str, tuple))

        # Filter widget attributes to avoid passing them to the callbacks
        attributes = self._filter_widget_attributes(kwargs)

        dynamic_width = kwargs.pop('dynamic_width', True)
        input_underline_vmargin = kwargs.pop('input_underline_vmargin', 0)
        prev_margin = kwargs.pop('previsualization_margin', 10)
        prev_width = kwargs.pop('previsualization_width', 3)

        widget = ColorInput(
            color_type=color_type,
            colorinput_id=color_id,
            cursor_color=self._theme.cursor_color,
            cursor_switch_ms=self._theme.cursor_switch_ms,
            dynamic_width=dynamic_width,
            hex_format=hex_format,
            input_separator=input_separator,
            input_underline=input_underline,
            input_underline_vmargin=input_underline_vmargin,
            onchange=onchange,
            onreturn=onreturn,
            onselect=onselect,
            prev_margin=prev_margin,
            prev_width_factor=prev_width,
            title=title,
            **kwargs
        )

        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)
        widget.set_default_value(default)

        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

SCROLLAREA
ScrollArea class to manage scrolling in Menu.
"""

__all__ = [

    # Main class
    'ScrollArea',

    # Utils
    'get_scrollbars_from_position'

]

import pygame
import pygame_menu

from itertools import product
from pygame_menu._base import Base
from pygame_menu._decorator import Decorator
from pygame_menu.locals import POSITION_SOUTHEAST, POSITION_SOUTHWEST, POSITION_WEST, \
    POSITION_NORTHEAST, POSITION_NORTHWEST, POSITION_CENTER, POSITION_EAST, \
    POSITION_NORTH, ORIENTATION_HORIZONTAL, ORIENTATION_VERTICAL, \
    SCROLLAREA_POSITION_BOTH_HORIZONTAL, POSITION_SOUTH, SCROLLAREA_POSITION_FULL, \
    SCROLLAREA_POSITION_BOTH_VERTICAL, SCROLLAREA_POSITION_NONE
from pygame_menu.utils import make_surface, assert_color, assert_position, \
    assert_orientation, get_finger_pos
from pygame_menu.widgets import ScrollBar

from pygame_menu._types import Union, NumberType, Tuple, List, Dict, Tuple2NumberType, \
    CursorInputType, Optional, Tuple2IntType, NumberInstance, ColorInputType, \
    EventVectorType, EventType, VectorInstance, StringVector, Any


def get_scrollbars_from_position(
    position: str
) -> Union[str, Tuple[str, str], Tuple[str, str, str, str]]:
    """
    Return the scrollbars from the given position.

    Raises ``ValueError`` if invalid position.

    :param position: Position
    :return: Scrollbars
    """
    if position in (POSITION_EAST, POSITION_EAST, POSITION_WEST, POSITION_NORTH):
        return position
    elif position == POSITION_NORTHWEST:
        return POSITION_NORTH, POSITION_WEST
    elif position == POSITION_NORTHEAST:
        return POSITION_NORTH, POSITION_EAST
    elif position == POSITION_SOUTHWEST:
        return POSITION_SOUTH, POSITION_WEST
    elif position == POSITION_SOUTHEAST:
        return POSITION_SOUTH, POSITION_EAST
    elif position == SCROLLAREA_POSITION_FULL:
        return POSITION_SOUTH, POSITION_EAST, POSITION_WEST, POSITION_NORTH
    elif position == SCROLLAREA_POSITION_BOTH_HORIZONTAL:
        return POSITION_SOUTH, POSITION_NORTH
    elif position == SCROLLAREA_POSITION_BOTH_VERTICAL:
        return POSITION_EAST, POSITION_WEST
    elif position == SCROLLAREA_POSITION_NONE:
        return ''
    elif position == POSITION_CENTER:
        raise ValueError('cannot init scrollbars from center position')
    else:
        raise ValueError('unknown ScrollArea position')


DEFAULT_SCROLLBARS = get_scrollbars_from_position(POSITION_SOUTHEAST)


class ScrollArea(Base):
    """
    The ScrollArea class provides a scrolling view managing up to 4 scroll bars.

    A scroll area is used to display the contents of a child surface (``world``).
    If the surface exceeds the size of the drawing surface, the view provide
    scroll bars so that the entire area of the child surface can be viewed.

    .. note::

        See :py:mod:`pygame_menu.locals` for valid ``scrollbars`` and
        ``shadow_position`` values.

    .. note::

        ScrollArea cannot be copied or deep-copied.

    :param area_width: Width of scrollable area in px
    :param area_height: Height of scrollable area in px
    :param area_color: Background color, it can be a color or an image
    :param border_color: Border color
    :param border_width: Border width in px
    :param controls_joystick: Use joystick events
    :param controls_keyboard: Use keyboard events
    :param controls_mouse: Use mouse events
    :param controls_touchscreen: Use touchscreen events
    :param extend_x: Px to extend the surface on x-axis in px from left. Recommended use only within Menus
    :param extend_y: Px to extend the surface on y-axis in px from top. Recommended use only within Menus
    :param menubar: Menubar for style compatibility. ``None`` if ScrollArea is not used within a Menu (for example, in Frames)
    :param parent_scrollarea: Parent ScrollArea if the new one is added within another area
    :param scrollarea_id: Scrollarea ID
    :param scrollbar_color: Scrollbars color
    :param scrollbar_cursor: Scrollbar cursor
    :param scrollbar_slider_color: Color of the sliders
    :param scrollbar_slider_hover_color: Color of the slider if hovered or clicked
    :param scrollbar_slider_pad: Space between slider and scrollbars borders in px
    :param scrollbar_thick: Scrollbar thickness in px
    :param scrollbars: Positions of the scrollbars. See :py:mod:`pygame_menu.locals`
    :param shadow: Indicate if a shadow is drawn on each scrollbar
    :param shadow_color: Color of the shadow of each scrollbar
    :param shadow_offset: Offset of the scrollbar shadow in px
    :param shadow_position: Position of the scrollbar shadow. See :py:mod:`pygame_menu.locals`
    :param world: Surface to draw and scroll
    """
    _area_color: Optional[Union[ColorInputType, 'pygame_menu.BaseImage']]
    _border_color: Optional[Union[ColorInputType, 'pygame_menu.BaseImage']]
    _border_tiles: List['pygame.Surface']
    _border_tiles_size: Tuple2IntType
    _border_width: int
    _bg_surface: Optional['pygame.Surface']
    _decorator: 'Decorator'
    _extend_x: int
    _extend_y: int
    _menu: Optional['pygame_menu.Menu']
    _menubar: 'pygame_menu.widgets.MenuBar'
    _parent_scrollarea: 'ScrollArea'
    _rect: 'pygame.Rect'
    _scrollbar_positions: Tuple[str, ...]
    _scrollbars: List['ScrollBar']
    _scrollbars_props: Tuple[Any, ...]
    _translate: Tuple2IntType
    _view_rect: 'pygame.Rect'
    _world: 'pygame.Surface'

    def __init__(
        self,
        area_width: int,
        area_height: int,
        area_color: Optional[Union[ColorInputType, 'pygame_menu.BaseImage']] = None,
        border_color: Optional[Union[ColorInputType, 'pygame_menu.BaseImage']] = None,
        border_width: int = 0,
        controls_joystick: bool = True,
        controls_keyboard: bool = True,
        controls_mouse: bool = True,
        controls_touchscreen: bool = True,
        extend_x: int = 0,
        extend_y: int = 0,
        menubar: Optional['pygame_menu.widgets.MenuBar'] = None,
        parent_scrollarea: Optional['ScrollArea'] = None,
        scrollarea_id: str = '',
        scrollbar_color: ColorInputType = (235, 235, 235),
        scrollbar_cursor: CursorInputType = None,
        scrollbar_slider_color: ColorInputType = (200, 200, 200),
        scrollbar_slider_hover_color: ColorInputType = (180, 180, 180),
        scrollbar_slider_pad: NumberType = 0,
        scrollbar_thick: int = 20,
        scrollbars: StringVector = DEFAULT_SCROLLBARS,
        shadow: bool = False,
        shadow_color: ColorInputType = (0, 0, 0),
        shadow_offset: int = 2,
        shadow_position: str = POSITION_SOUTHEAST,
        world: Optional['pygame.Surface'] = None
    ) -> None:
        super(ScrollArea, self).__init__(object_id=scrollarea_id)

        assert isinstance(area_height, int)
        assert isinstance(area_width, int)
        assert isinstance(border_width, int)
        assert isinstance(controls_joystick, bool)
        assert isinstance(controls_keyboard, bool)
        assert isinstance(controls_mouse, bool)
        assert isinstance(controls_touchscreen, bool)
        assert isinstance(extend_x, int)
        assert isinstance(extend_y, int)
        assert isinstance(scrollbar_slider_pad, NumberInstance)
        assert isinstance(scrollbar_thick, int)
        assert isinstance(shadow, bool)
        assert isinstance(shadow_offset, int)
        assert isinstance(world, (pygame.Surface, type(None)))

        if area_color is not None and not isinstance(area_color, pygame_menu.BaseImage):
            area_color = assert_color(area_color)
        if border_color is not None and not isinstance(border_color, pygame_menu.BaseImage):
            border_color = assert_color(border_color)

        # Create tiles
        if isinstance(border_color, pygame_menu.BaseImage):
            iw, ih = border_color.get_size()
            tw, th = iw // 3, ih // 3
            self._border_tiles_size = tw, th
            self._border_tiles = [
                border_color.subsurface((x, y, tw, th))
                for x, y in product(range(0, iw, tw), range(0, ih, th))
            ]

        scrollbar_color = assert_color(scrollbar_color)
        scrollbar_slider_color = assert_color(scrollbar_slider_color)
        shadow_color = assert_color(shadow_color)

        assert_position(shadow_position)

        assert area_width > 0 and area_height > 0, 'area size must be greater than zero'

        assert isinstance(scrollbars, (str, VectorInstance))
        unique_scrolls = []
        if isinstance(scrollbars, str):
            unique_scrolls.append(scrollbars)
        else:
            for s in scrollbars:
                if s not in unique_scrolls:
                    unique_scrolls.append(s)

        # Remove none position
        if '' in unique_scrolls:
            unique_scrolls.pop(unique_scrolls.index(''))

        self._area_color = area_color
        self._border_color = border_color
        self._border_width = border_width
        self._bg_surface = None
        self._decorator = Decorator(self)
        self._scrollbar_positions = tuple(unique_scrolls)  # Ensure unique
        self._translate = (0, 0)
        self._world = world

        self._extend_x = extend_x
        self._extend_y = extend_y
        self._menubar = menubar

        self._scrollbars_props = (scrollbar_color, scrollbar_thick, scrollbar_slider_color,
                                  scrollbar_slider_hover_color, scrollbar_slider_pad,
                                  scrollbar_cursor, shadow, shadow_color, shadow_position,
                                  shadow_offset, controls_joystick, controls_mouse,
                                  controls_touchscreen, controls_keyboard)
        self.set_parent_scrollarea(parent_scrollarea)
        self.create_rect(area_width, area_height)

        # Menu reference
        self._menu = None

    def create_rect(self, width: int, height: int) -> None:
        """
        Create rect object.

        :param width: Area width
        :param height: Area height
        """
        assert isinstance(width, int)
        assert isinstance(height, int)
        self._rect = pygame.Rect(0, 0, int(width), int(height))
        self._scrollbars = []
        self._view_rect = self.get_view_rect()

        # Unpack properties
        (scrollbar_color, scrollbar_thick, scrollbar_slider_color,
         scrollbar_slider_hover_color, scrollbar_slider_pad,
         scrollbar_cursor, shadow, shadow_color, shadow_position,
         shadow_offset, controls_joystick, controls_mouse, controls_touchscreen,
         controls_keyboard) = self._scrollbars_props

        for pos in self._scrollbar_positions:
            assert_position(pos)

            if pos == POSITION_EAST or pos == POSITION_WEST:
                sbar = ScrollBar(
                    length=self._view_rect.height,
                    onchange=self._on_vertical_scroll,
                    orientation=ORIENTATION_VERTICAL,
                    page_ctrl_color=scrollbar_color,
                    page_ctrl_thick=scrollbar_thick,
                    slider_color=scrollbar_slider_color,
                    slider_hover_color=scrollbar_slider_hover_color,
                    slider_pad=scrollbar_slider_pad,
                    values_range=(0, max(1, self.get_hidden_height()))
                )

            else:
                sbar = ScrollBar(
                    length=self._view_rect.width,
                    onchange=self._on_horizontal_scroll,
                    page_ctrl_color=scrollbar_color,
                    page_ctrl_thick=scrollbar_thick,
                    slider_color=scrollbar_slider_color,
                    slider_hover_color=scrollbar_slider_hover_color,
                    slider_pad=scrollbar_slider_pad,
                    values_range=(0, max(1, self.get_hidden_width()))
                )

            sbar.set_shadow(
                enabled=shadow,
                color=shadow_color,
                position=shadow_position,
                offset=shadow_offset
            )
            sbar.set_controls(
                joystick=controls_joystick,
                mouse=controls_mouse,
                touchscreen=controls_touchscreen,
                keyboard=controls_keyboard
            )
            sbar.set_cursor(cursor=scrollbar_cursor)
            sbar.set_scrollarea(self)
            sbar.configured = True
            sbar.hide()

            self._scrollbars.append(sbar)

        self._apply_size_changes()

    def _make_background_surface(self) -> None:
        """
        Create background surface.
        """
        # If bg surface is created, and it's the same size
        if (
            self._bg_surface is not None and
            self._bg_surface.get_width() == self._rect.width and
            self._bg_surface.get_height() == self._rect.height
        ):
            return

        # Make surface
        self._bg_surface = make_surface(width=self._rect.width + self._extend_x, height=self._rect.height + self._extend_y)
        rect = self._bg_surface.get_rect()
        if self._area_color is not None:
            if isinstance(self._area_color, pygame_menu.BaseImage):
                self._area_color.draw(surface=self._bg_surface, area=rect)
            else:
                self._bg_surface.fill(assert_color(self._area_color))

    def update_area_color(
        self,
        color: Optional[Union[ColorInputType, 'pygame_menu.BaseImage']]
    ) -> 'ScrollArea':
        """
        Updates area color (background).

        :param color: Color
        :return: Self reference
        """
        self._area_color = color
        self._bg_surface = None
        self._make_background_surface()
        return self

    def set_parent_scrollarea(self, parent: Optional['ScrollArea']) -> 'ScrollArea':
        """
        Set parent ScrollArea.

        :param parent: Parent ScrollArea
        """
        assert isinstance(parent, (ScrollArea, type(None)))
        assert parent != self, 'parent scrollarea cannot be set as itself'
        self._parent_scrollarea = parent
        return self

    def get_parent(self) -> Optional['ScrollArea']:
        """
        Return the parent ScrollArea.

        :return: Parent ScrollArea object
        """
        return self._parent_scrollarea

    def get_depth(self) -> int:
        """
        Return the depth of the ScrollArea (how many parents do it has recursively).

        :return: Depth's number
        """
        parent = self._parent_scrollarea
        count = 0
        if parent is not None:
            while True:
                if parent is None:
                    break
                count += 1
                parent = parent._parent_scrollarea
        return count

    def __copy__(self) -> 'ScrollArea':
        """
        Copy method.

        :return: Raises copy exception
        """
        raise _ScrollAreaCopyException('ScrollArea class cannot be copied')

    def __deepcopy__(self, memodict: Dict) -> 'ScrollArea':
        """
        Deep-copy method.

        :param memodict: Memo dict
        :return: Raises copy exception
        """
        raise _ScrollAreaCopyException('ScrollArea class cannot be copied')

    def force_menu_surface_update(self) -> 'ScrollArea':
        """
        Forces menu surface update after next rendering call.

        .. note::

            This method is expensive, as menu surface update forces re-rendering
            of all widgets (because them can change in size, position, etc...).

        :return: Self reference
        """
        if self._menu is not None:
            self._menu._widgets_surface_need_update = True
        return self

    def force_menu_surface_cache_update(self) -> 'ScrollArea':
        """
        Forces menu surface cache to update after next drawing call.
        This also updates widget decoration.

        .. note::

            This method only updates the surface cache, without forcing re-rendering
            of all Menu widgets as
            :py:meth:`pygame_menu.widgets.core.widget.Widget.force_menu_surface_update`
            does.

        :return: Self reference
        """
        if self._menu is not None:
            self._menu._widget_surface_cache_need_update = True
            self._decorator.force_cache_update()
        return self

    def _apply_size_changes(self) -> None:
        """
        Apply size changes to scrollbar.
        """
        self._view_rect = self.get_view_rect()

        for sbar in self._scrollbars:
            pos = self._scrollbar_positions[self._scrollbars.index(sbar)]

            d_size, dx, dy = 0, 0, 0
            if self._menubar is not None:
                d_size, (dx, dy) = self._menubar.get_scrollbar_style_change(pos)

            if pos == POSITION_WEST:
                sbar.set_position(x=self._view_rect.left - sbar.get_thickness() + dx, y=self._view_rect.top + dy)
            elif pos == POSITION_EAST:
                sbar.set_position(x=self._view_rect.right + dx, y=self._view_rect.top + dy)
            elif pos == POSITION_NORTH:
                sbar.set_position(x=self._view_rect.left + dx, y=self._view_rect.top - sbar.get_thickness() + dy)
            elif pos == POSITION_SOUTH:  # South
                sbar.set_position(x=self._view_rect.left + dx, y=self._view_rect.bottom + dy)
            else:
                raise ValueError('unknown position, only west, east, north, and'
                                 'south are allowed')

            if pos in (POSITION_NORTH, POSITION_SOUTH):
                if self.get_hidden_width() != 0:
                    sbar.set_length(self._view_rect.width + d_size)
                    sbar.set_maximum(self.get_hidden_width())
                    sbar.set_page_step(self._view_rect.width * self.get_hidden_width() /
                                       (self._view_rect.width + self.get_hidden_width()))
                    sbar.show()
                else:
                    sbar.hide()

            elif pos in (POSITION_EAST, POSITION_WEST):
                if self.get_hidden_height() != 0:
                    sbar.set_length(self._view_rect.height + d_size)
                    sbar.set_maximum(self.get_hidden_height())
                    sbar.set_page_step(self._view_rect.height * self.get_hidden_height() /
                                       (self._view_rect.height + self.get_hidden_height()))
                    sbar.show()
                else:
                    sbar.hide()

    def draw(self, surface: 'pygame.Surface') -> 'ScrollArea':
        """
        Draw the ScrollArea.

        :param surface: Surface to render the area
        :return: Self reference
        """
        if not self._world:
            return self

        # Background surface already has previous decorators
        if self._area_color is not None:
            self._make_background_surface()
            surface.blit(self._bg_surface, (self._rect.x - self._extend_x, self._rect.y - self._extend_y))

        # Draw world surface
        # noinspection PyTypeChecker
        surface.blit(self._world, self._view_rect.topleft, (self.get_offsets(), self._view_rect.size))

        # Then draw scrollbars
        for sbar in self._scrollbars:
            if not sbar.is_visible():
                continue
            if sbar.get_orientation() == ORIENTATION_HORIZONTAL:
                if self.get_hidden_width():
                    sbar.draw(surface)
            else:
                if self.get_hidden_height():
                    sbar.draw(surface)

        # Draw post decorator
        self._decorator.draw_post(surface)

        # Create border
        if isinstance(self._border_color, pygame_menu.BaseImage):  # Image
            tw, th = self._border_tiles_size
            border_rect = pygame.Rect(
                int(self._rect.x - tw),
                int(self._rect.y - th),
                int(self._rect.width + 2 * tw),
                int(self._rect.height + 2 * th)
            )

            surface_blit = surface.blit
            (
                tile_nw,
                tile_w,
                tile_sw,
                tile_n,
                tile_c,
                tile_s,
                tile_ne,
                tile_e,
                tile_se,
            ) = self._border_tiles
            left, top = self._rect.topleft
            left -= tw
            top -= th

            # draw top and bottom tiles
            area: Optional[Tuple[int, int, int, int]]

            for x in range(border_rect.left, border_rect.right, tw):
                if x + tw >= border_rect.right:
                    area = 0, 0, tw - (x + border_rect.right), th
                else:
                    area = None
                surface_blit(tile_n, (x, top), area)
                surface_blit(tile_s, (x, border_rect.bottom - th), area)

            # draw left and right tiles
            for y in range(border_rect.top, border_rect.bottom, th):
                if y + th >= border_rect.bottom:
                    area = 0, 0, tw, th - (y + border_rect.bottom)
                else:
                    area = None
                surface_blit(tile_w, (left, y), area)
                surface_blit(tile_e, (border_rect.right - tw, y), area)

            # draw corners
            surface_blit(tile_nw, (left, top))
            surface_blit(tile_sw, (left, border_rect.bottom - th))
            surface_blit(tile_ne, (border_rect.right - tw, top))
            surface_blit(tile_se, (border_rect.right - tw, border_rect.bottom - th))

        else:  # Color
            if self._border_width == 0 or self._border_color is None:
                return self
            border_rect = pygame.Rect(
                int(self._rect.x - self._border_width),
                int(self._rect.y - self._border_width),
                int(self._rect.width + 2 * self._border_width),
                int(self._rect.height + 2 * self._border_width)
            )
            pygame.draw.rect(
                surface,
                self._border_color,
                border_rect,
                self._border_width
            )

        return self

    def get_border_size(self) -> Tuple2IntType:
        """
        Return the border size (width, height) in px.

        :return: Border width, height
        """
        if isinstance(self._border_color, pygame_menu.BaseImage):  # Image
            return self._border_tiles_size
        else:  # Color
            if self._border_color is None:
                return 0, 0
            return self._border_width, self._border_width

    def get_hidden_width(self) -> int:
        """
        Return the total width out of the bounds of the viewable area.
        Zero is returned if the world width is lower than the viewable area.

        :return: Hidden width in px
        """
        if not self._world:
            return 0
        return int(max(0, self._world.get_width() - self._view_rect.width))

    def get_hidden_height(self) -> int:
        """
        Return the total height out of the bounds of the viewable area.
        Zero is returned if the world height is lower than the viewable area.

        :return: Hidden height in px
        """
        if not self._world:
            return 0
        return int(max(0, self._world.get_height() - self._view_rect.height))

    def get_offsets(self) -> Tuple2IntType:
        """
        Return the offset introduced by the scrollbars in the world.

        :return: ScrollArea offset on x-axis and y-axis (x, y)
        """
        offsets = [0, 0]
        for sbar in self._scrollbars:
            if not sbar.is_visible():
                continue
            if sbar.get_orientation() == ORIENTATION_HORIZONTAL:
                if self.get_hidden_width():
                    offsets[0] = sbar.get_value()  # Cannot add as each scrollbar can only affect 1 axis only
            else:
                if self.get_hidden_height():
                    offsets[1] = sbar.get_value()
        return offsets[0], offsets[1]

    def get_rect(self, to_real_position: bool = False) -> 'pygame.Rect':
        """
        Return the :py:class:`pygame.Rect` object of the ScrollArea.

        :param to_real_position: Get real position fof the scroll area
        :return: Pygame.Rect object
        """
        rect = self._rect.copy()
        if to_real_position:
            rect = self.to_real_position(rect)
        return rect

    def get_scrollbar_thickness(self, orientation: str, visible: bool = True) -> int:
        """
        Return the scroll thickness of the area. If it's hidden return zero.

        :param orientation: Orientation of the scroll. See :py:mod:`pygame_menu.locals`
        :param visible: If ``True`` returns the real thickness depending on if it is visible or not
        :return: Thickness in px
        """
        assert_orientation(orientation)
        assert isinstance(visible, bool)

        if visible:
            total = 0
            for sbar in self._scrollbars:
                if sbar.get_orientation() == orientation and sbar.is_visible():
                    total += sbar.get_thickness()
            return total

        if orientation == ORIENTATION_HORIZONTAL:
            return int(self._rect.height - self._view_rect.height)
        elif orientation == ORIENTATION_VERTICAL:
            return int(self._rect.width - self._view_rect.width)

    def get_world_rect(self, absolute: bool = False) -> 'pygame.Rect':
        """
        Return the world rect.

        :param absolute: To absolute position
        :return: World rect object
        """
        rect = self._world.get_rect()
        if absolute:
            rect = self.to_absolute_position(rect)
        return rect

    def get_view_rect(self) -> 'pygame.Rect':
        """
        Subtract width of scrollbars from area with the given size and return
        the viewable area.

        The viewable area depends on the world size, because scroll bars may or
        may not be displayed.

        :return: View rect object
        """
        rect = pygame.Rect(self._rect)

        # No scrollbar: area is large enough to display world
        if not self._world or (self._world.get_width() <= self._rect.width
                               and self._world.get_height() <= self._rect.height):
            return rect

        # All scrollbars: the world is too large
        if self._world.get_height() > self._rect.height and self._world.get_width() > self._rect.width:
            for sbar in self._scrollbars:
                if not sbar.is_visible():
                    continue
                pos = self._scrollbar_positions[self._scrollbars.index(sbar)]
                thk = sbar.get_thickness()
                if pos == POSITION_WEST:
                    rect.left += thk
                    rect.width -= thk
                elif pos == POSITION_EAST:
                    rect.width -= thk
                elif pos == POSITION_NORTH:
                    rect.top += thk
                    rect.height -= thk
                elif pos == POSITION_SOUTH:
                    rect.height -= thk
            return rect

        # Calculate the maximum variations introduced by the scrollbars
        bars_total_width = 0
        bars_total_height = 0
        for sbar in self._scrollbars:
            if not sbar.is_visible():
                continue
            pos = self._scrollbar_positions[self._scrollbars.index(sbar)]
            thk = sbar.get_thickness()
            if pos in (POSITION_NORTH, POSITION_SOUTH):
                bars_total_height += thk
            elif pos in (POSITION_WEST, POSITION_EAST):
                bars_total_width += thk

        if self._world.get_height() > self._rect.height:
            for sbar in self._scrollbars:
                if not sbar.is_visible():
                    continue
                pos = self._scrollbar_positions[self._scrollbars.index(sbar)]
                thk = sbar.get_thickness()
                if pos == POSITION_WEST:
                    rect.left += thk
                    rect.width -= thk
                elif pos == POSITION_EAST:
                    rect.width -= thk
                if self._world.get_width() > self._rect.width - bars_total_width:
                    if pos == POSITION_NORTH:
                        rect.top += thk
                        rect.height -= thk
                    elif pos == POSITION_SOUTH:
                        rect.height -= thk

        if self._world.get_width() > self._rect.width:
            for sbar in self._scrollbars:
                if not sbar.is_visible():
                    continue
                pos = self._scrollbar_positions[self._scrollbars.index(sbar)]
                thk = sbar.get_thickness()
                if pos == POSITION_NORTH:
                    rect.top += thk
                    rect.height -= thk
                elif pos == POSITION_SOUTH:
                    rect.height -= thk
                if self._world.get_height() > self._rect.height - bars_total_height:
                    if pos == POSITION_WEST:
                        rect.left += thk
                        rect.width -= thk
                    elif pos == POSITION_EAST:
                        rect.width -= thk

        return rect

    def hide_scrollbars(self, orientation: str, force: bool = True) -> 'ScrollArea':
        """
        Hide scrollbar from given orientation.

        :param orientation: Orientation. See :py:mod:`pygame_menu.locals`
        :param force: Force hide
        :return: Self reference
        """
        assert_orientation(orientation)
        for sbar in self._scrollbars:
            if sbar.get_orientation() == orientation:
                sbar.hide(force=force)
                if not force:
                    sbar.disable_visibility_force()
        self._apply_size_changes()
        return self

    def show_scrollbars(self, orientation: str, force: bool = True) -> 'ScrollArea':
        """
        Hide scrollbar from given orientation.

        :param orientation: Orientation. See :py:mod:`pygame_menu.locals`
        :param force: Force show
        :return: Self reference
        """
        assert_orientation(orientation)
        for sbar in self._scrollbars:
            if sbar.get_orientation() == orientation:
                sbar.show(force=force)
                if not force:
                    sbar.disable_visibility_force()
        self._apply_size_changes()
        return self

    def get_world_size(self) -> Tuple2IntType:
        """
        Return the world size.

        :return: Width, height in pixels
        """
        if self._world is None:
            return 0, 0
        return self._world.get_width(), self._world.get_height()

    def get_size(self, inner: bool = False) -> Tuple2IntType:
        """
        Return the area size.

        :param inner: If ``True`` returns the rect view area
        :return: Width, height in pixels
        """
        if inner:
            return self._view_rect.width, self._view_rect.height
        return self._rect.width, self._rect.height

    def mouse_is_over(self, view: bool = False) -> bool:
        """
        Return ``True`` if the mouse is placed over the ScrollArea.

        :param view: If ``True`` uses "view rect" instead of "rect"
        :return: ``True`` if the mouse is over the object
        """
        rect = self._view_rect if view else self._rect
        return bool(self.to_absolute_position(rect).collidepoint(*pygame.mouse.get_pos()))

    def _on_horizontal_scroll(self, value: NumberType) -> None:
        """
        Call when a horizontal scroll bar as changed to update the
        position of the opposite one if it exists.

        :param value: New position of the slider
        """
        for sbar in self._scrollbars:
            if (
                sbar.get_orientation() == ORIENTATION_HORIZONTAL and
                self.get_hidden_width() != 0 and
                sbar.get_value() != value
            ):
                sbar.set_value(value)

    def _on_vertical_scroll(self, value: NumberType) -> None:
        """
        Call when a vertical scroll bar as changed to update the
        position of the opposite one if it exists.

        :param value: New position of the slider
        """
        for sbar in self._scrollbars:
            if (
                sbar.get_orientation() == ORIENTATION_VERTICAL and
                self.get_hidden_height() != 0 and
                sbar.get_value() != value
            ):
                sbar.set_value(value)

    def get_parent_scroll_value_percentage(self, orientation: str) -> Tuple[float]:
        """
        Get percentage scroll values of scroll and parents; if ``0`` the scroll
        is at top/left, ``1`` bottom/right.

        :param orientation: Orientation. See :py:mod:`pygame_menu.locals`
        :return: Value from ``0`` to ``1`` as a tuple; first item is the current scrollarea
        """
        values = [self.get_scroll_value_percentage(orientation)]
        parent = self._parent_scrollarea
        if parent is not None:
            while True:  # Recursive
                if parent is None:
                    break
                values.append(parent.get_scroll_value_percentage(orientation))
                parent = parent._parent_scrollarea
        return tuple(values)

    def get_scroll_value_percentage(self, orientation: str) -> float:
        """
        Get the scroll value in percentage; if ``0`` the scroll is at top/left,
        ``1`` bottom/right.

        .. note::

            If ScrollArea does not contain such orientation scroll, ``-1`` is returned.

        :param orientation: Orientation. See :py:mod:`pygame_menu.locals`
        :return: Value from ``0`` to ``1``
        """
        assert_orientation(orientation)
        for sbar in self._scrollbars:
            if not sbar.is_visible():
                continue
            if sbar.get_orientation() == orientation:
                return sbar.get_value_percentage()
        return -1

    def scroll_to(self, orientation: str, value: NumberType) -> 'ScrollArea':
        """
        Scroll to position in terms of the percentage.

        :param orientation: Orientation. See :py:mod:`pygame_menu.locals`
        :param value: If ``0`` scrolls to top/left, ``1`` to bottom/right
        :return: Self reference
        """
        assert_orientation(orientation)
        assert isinstance(value, NumberInstance) and 0 <= value <= 1
        for sbar in self._scrollbars:
            if not sbar.is_visible():
                continue
            if sbar.get_orientation() == orientation:
                v_min, v_max = sbar.get_minmax()
                delta = v_max - v_min
                new_value = int(min(v_min + delta * float(value), v_max))
                sbar.set_value(new_value)
                break
        return self

    # noinspection PyTypeChecker
    def scroll_to_rect(
        self,
        rect: 'pygame.Rect',
        margin: Tuple2NumberType = (0, 0),
        scroll_parent: bool = True
    ) -> bool:
        """
        Ensure that the given rect is in the viewable area.

        :param rect: Rect in the world surface reference
        :param margin: Extra margin around the rect on x-axis and y-axis in px
        :param scroll_parent: If ``True`` parent scroll also scrolls to rect
        :return: Scrollarea scrolled to rect. If ``False`` the rect was already inside the visible area
        """
        # Check if visible
        if self.to_real_position(rect, visible=True).height == 0 and self._parent_scrollarea is not None and scroll_parent:
            self._parent_scrollarea.scroll_to_rect(self._parent_scrollarea.get_rect(), margin, scroll_parent)
            self._parent_scrollarea.scroll_to_rect(self.get_rect(), margin, scroll_parent)

        real_rect = self.to_real_position(rect)

        # Add margin to rect
        real_rect.x += margin[0]
        real_rect.y += margin[1]

        # Check rect is in viewable area
        sx = self.get_scrollbar_thickness(ORIENTATION_VERTICAL)
        sy = self.get_scrollbar_thickness(ORIENTATION_HORIZONTAL)
        view_rect = self.get_absolute_view_rect()
        if (
            view_rect.topleft[0] <= real_rect.topleft[0] + sx and
            view_rect.topleft[1] <= real_rect.topleft[1] + sy and
            view_rect.bottomright[0] + sx >= real_rect.bottomright[0] and
            view_rect.bottomright[1] + sy >= real_rect.bottomright[1]
        ):
            return False

        for sbar in self._scrollbars:
            if not sbar.is_visible():
                continue
            if sbar.get_orientation() == ORIENTATION_HORIZONTAL and self.get_hidden_width():
                shortest_move = min(real_rect.left - view_rect.left, real_rect.right - view_rect.right, key=abs)
                value = min(sbar.get_maximum(), sbar.get_value() + shortest_move)
                value = max(sbar.get_minimum(), value)
                sbar.set_value(value)
            if sbar.get_orientation() == ORIENTATION_VERTICAL and self.get_hidden_height():
                shortest_move = min(real_rect.bottom - view_rect.bottom, real_rect.top - view_rect.top, key=abs)
                value = min(sbar.get_maximum(), sbar.get_value() + shortest_move)
                value = max(sbar.get_minimum(), value)
                sbar.set_value(value)

        if self._parent_scrollarea is not None and scroll_parent:
            self._parent_scrollarea.scroll_to_rect(rect, margin, scroll_parent)

        return True

    def set_position(self, x: int, y: int) -> 'ScrollArea':
        """
        Set the position.

        :param x: X position
        :param y: Y position
        :return: Self reference
        """
        self._rect.x = x + self._extend_x + self._translate[0]
        self._rect.y = y + self._extend_y + self._translate[1]
        self._apply_size_changes()
        return self

    def get_position(self) -> Tuple2IntType:
        """
        Return the ScrollArea position.

        :return: X, Y position in px
        """
        return self._rect.x, self._rect.y

    def get_widget_position_relative_to_view_rect(
        self,
        widget: 'pygame_menu.widgets.Widget'
    ) -> Tuple2NumberType:
        """
        Get widget position relative to view rect on x-axis and y-axis. On each axis,
        the relative position goes from ``-inf`` to ``+inf``. If between (0, 1) the
        widget is inside the view rect.

        .. note::

            Only top-left widget position is checked.

        :param widget: Widget to check the position
        :return: Relative position to view rect on x-axis and y-axis
        """
        assert widget.get_scrollarea() == self, \
            '{0} scrollarea {1} is different than current {2}' \
                .format(widget, widget.get_scrollarea().get_class_id(), self.get_class_id())
        wx, wy = widget.get_position()
        view_rect = self.get_view_rect()
        vx, vy = view_rect.width, view_rect.height
        offx, offy = self.get_offsets()
        return (wx - offx) / vx, (wy - offy) / vy

    def translate(self, x: NumberType, y: NumberType) -> 'ScrollArea':
        """
        Translate on x-axis and y-axis (x, y) in px.

        :param x: X translation in px
        :param y: Y translation in px
        :return: Self reference
        """
        assert isinstance(x, NumberInstance)
        assert isinstance(y, NumberInstance)
        self._rect.x -= self._translate[0]
        self._rect.y -= self._translate[1]
        self._translate = (x, y)
        self._rect.x += x
        self._rect.y += y
        self._apply_size_changes()
        return self

    def get_translate(self) -> Tuple2IntType:
        """
        Get object translation on both axis.

        :return: Translation on x-axis and y-axis (x, y) in px
        """
        return self._translate

    def set_world(self, surface: 'pygame.Surface') -> 'ScrollArea':
        """
        Update the scrolled surface.

        :param surface: New world surface
        :return: Self reference
        """
        self._world = surface
        self._apply_size_changes()
        return self

    def get_world(self) -> Optional['pygame.Surface']:
        """
        Return the world surface area.

        .. warning::

            Use with caution.

        :return: World surface. ``None`` if it has not been set yet
        """
        return self._world

    def get_parent_position(self) -> Tuple2IntType:
        """
        Return parent ScrollArea position.

        :return: Position on x, y-axis in px
        """
        if self._parent_scrollarea is not None:
            px, py = self._parent_scrollarea.get_position()
            ox, oy = self._parent_scrollarea.get_offsets()
            par_x, par_y = 0, 0
            if self._parent_scrollarea.get_parent() is not None:
                par_x, par_y = self._parent_scrollarea.get_parent_position()
            return px - ox + par_x, py - oy + par_y
        return 0, 0

    def to_absolute_position(self, virtual: 'pygame.Rect') -> 'pygame.Rect':
        """
        Return the absolute position of a rect within the ScrollArea. Absolute
        position is concerning the parent ScrollArea. If ``None``, the rect is
        not changed at all.

        .. note::

            Absolute position must be used if desired to get the widget position
            outside a scrolled area status, for example the view rect, or the
            scrollbars.

        :param virtual: Rect in the world surface reference
        :return: Rect in absolute position
        """
        rect = pygame.Rect(virtual)
        parent_position = self.get_parent_position()
        rect.x += parent_position[0]
        rect.y += parent_position[1]
        return rect

    def get_absolute_view_rect(self) -> 'pygame.Rect':
        """
        Return the ScrollArea absolute view rect clipped if it is not visible by
        its parent ScrollArea.

        :return: Clipped absolute view rect
        """
        view_rect_absolute = self.to_absolute_position(self._view_rect)
        if self._parent_scrollarea is not None:
            parent = self._parent_scrollarea
            if parent is not None:
                while True:  # Recursive
                    if parent is None:
                        break
                    view_rect_absolute = parent.get_absolute_view_rect().clip(view_rect_absolute)
                    parent = parent._parent_scrollarea
        return view_rect_absolute

    def to_real_position(
        self,
        virtual: Union['pygame.Rect', Tuple2NumberType],
        visible: bool = False
    ) -> Union['pygame.Rect', Tuple2IntType]:
        """
        Return the real position/Rect according to the ScrollArea origin of a
        position/Rect in the world surface reference.

        .. note::

            Real position must be used if desired to get the widget position within
            a scrolled area status.

        :param virtual: Position/Rect in the world surface reference
        :param visible: If a ``virtual`` is Rect object, return only the visible width/height
        :return: Real rect or real position
        """
        assert isinstance(virtual, (pygame.Rect, VectorInstance))
        offsets = self.get_offsets()
        parent_position = self.get_parent_position()

        if isinstance(virtual, pygame.Rect):
            rect = pygame.Rect(virtual)  # virtual.copy() should also work
            rect.x = virtual.x + self._rect.x - offsets[0] + parent_position[0]
            rect.y = virtual.y + self._rect.y - offsets[1] + parent_position[1]
            if visible:
                return self.get_absolute_view_rect().clip(rect)  # Visible width and height
            return rect

        x_coord = self._rect.x + virtual[0] - offsets[0] + parent_position[0]
        y_coord = self._rect.y + virtual[1] - offsets[1] + parent_position[1]
        return int(x_coord), int(y_coord)

    def to_world_position(
        self,
        real: Union['pygame.Rect', Tuple2NumberType]
    ) -> Union['pygame.Rect', Tuple2IntType]:
        """
        Return the position/Rect in the world surface reference of a real
        position/Rect according to the ScrollArea origin.

        .. note::

            Virtual position must be used if desired to get the widget position
            within a scrolled area status.

        :param real: Position/Rect according ScrollArea origin
        :return: Rect in world or position in world
        """
        assert isinstance(real, (pygame.Rect, VectorInstance))
        offsets = self.get_offsets()
        parent_position = self.get_parent_position()

        if isinstance(real, pygame.Rect):
            rect = pygame.Rect(real)
            rect.x = real.x - self._rect.x + offsets[0] - parent_position[0]
            rect.y = real.y - self._rect.y + offsets[1] - parent_position[1]
            return rect

        x_coord = real[0] - self._rect.x + offsets[0] - parent_position[0]
        y_coord = real[1] - self._rect.y + offsets[1] - parent_position[1]
        return int(x_coord), int(y_coord)

    def is_scrolling(self) -> bool:
        """
        Return ``True`` if the user is scrolling.

        :return: ``True`` if user scrolls
        """
        scroll = False
        for sbar in self._scrollbars:
            scroll = scroll or sbar.scrolling
        return scroll

    def update(self, events: EventVectorType) -> bool:
        """
        Called by end user to update scroll state.

        :param events: List of pygame events
        :return: ``True`` if updated
        """
        updated = [False, False]
        for sbar in self._scrollbars:
            if not sbar.is_visible():
                continue
            if self.get_hidden_width() and not updated[0] and sbar.get_orientation() == ORIENTATION_HORIZONTAL:
                updated[0] = sbar.update(events)
            elif self.get_hidden_height() and not updated[1] and sbar.get_orientation() == ORIENTATION_VERTICAL:
                updated[1] = sbar.update(events)
        return updated[0] or updated[1]

    def set_menu(self, menu: 'pygame_menu.Menu') -> 'ScrollArea':
        """
        Set the Menu reference.

        :param menu: Menu object
        :return: Self reference
        """
        self._menu = menu
        for sbar in self._scrollbars:
            sbar.set_menu(menu)
        return self

    def get_menu(self) -> Optional['pygame_menu.Menu']:
        """
        Return the Menu reference (if exists).

        :return: Menu reference
        """
        return self._menu

    def collide(
        self,
        widget: Union['pygame_menu.widgets.Widget', 'pygame.Rect'],
        event: EventType
    ) -> bool:
        """
        If user event collides a widget within the ScrollArea respect to the
        relative position.

        :param widget: Widget or rect
        :param event: Pygame event
        :return: ``True`` if collide
        """
        if not isinstance(widget, pygame.Rect):
            widget_rect = widget.get_rect(to_real_position=True)
        else:
            widget_rect = widget
        return bool(widget_rect.collidepoint(*get_finger_pos(self._menu, event)))

    def get_decorator(self) -> 'Decorator':
        """
        Return the ScrollArea decorator API.

        .. note::

            Menu drawing order:

            1. Menu background color/image
            2. Menu ``prev`` decorator
            3. Menu **ScrollArea** ``prev`` decorator
            4. Menu **ScrollArea** widgets
            5. Menu **ScrollArea** ``post`` decorator
            6. Menu title
            7. Menu ``post`` decorator

        :return: Decorator API
        """
        return self._decorator


class _ScrollAreaCopyException(Exception):
    """
    If user tries to copy a ScrollArea.
    """
    pass
